###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Extension of DIRAC SiteDirector.

Simply defines what to send.
"""

from DIRAC import S_OK
from DIRAC.WorkloadManagementSystem.Agent.SiteDirector import SiteDirector as DIRACSiteDirector


class SiteDirector(DIRACSiteDirector):
    """Simple extension of the DIRAC site director to send LHCb specific pilots
    (with a custom list of commands)"""

    def beginExecution(self):
        """just simple redefinition."""
        res = DIRACSiteDirector.beginExecution(self)
        if not res["OK"]:
            return res

        self.lbRunOnly = self.am_getOption("lbRunOnly", False)

        return S_OK()

    def _getPilotOptions(self, queue):
        """Adding LHCb specific options."""
        pilotOptions = super()._getPilotOptions(queue)

        if self.lbRunOnly:
            pilotOptions.append("-o lbRunOnly")

        return pilotOptions



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""An agent to check for MCSimulation productions that have undergone the
testing phase. Productions that have the status Idle and are also in the table
StoredJobDescription have undergone testing. A report is created by the agent
from the results of the test phase and emailed to the Production Manager.

Author: Simon Bidwell
"""

from DIRAC import S_OK, S_ERROR

from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Workflow.Workflow import fromXMLString
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getUserOption

from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.ProductionManagementSystem.Client.Production import Production
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.Workflow.Modules.ModulesUtilities import getEventsToProduce, getCPUNormalizationFactorAvg

AGENT_NAME = "Transformation/MCSimulationTestingAgent"


class MCSimulationTestingAgent(AgentModule):
    """An agent to check for MCSimulation productions that have undergone the
    testing phase.

    Productions that have the status Idle and are also in the table
    StoredJobDescription have undergone testing. A report is created by
    the agent from the results of the test phase and emailed to the
    Production Manager
    """

    def __init__(self, *args, **kwargs):
        """c'tor."""
        AgentModule.__init__(self, *args, **kwargs)
        self.transClient = None
        self.bkClient = None
        self.notifyClient = None
        self.operations = None

        self.failedTransIDs = []

    def initialize(self):
        self.transClient = TransformationClient()
        self.bkClient = BookkeepingClient()
        self.notifyClient = NotificationClient()
        self.operations = Operations()

        self.email = self.am_getOption("MailTo", "")

        return S_OK()

    def execute(self):
        # get all the idle transformations
        extendableTTypes = Operations().getValue("Transformations/ExtendableTransfTypes", ["MCSimulation"])
        res = self.transClient.getTransformations(condDict={"Status": "Idle", "Type": extendableTTypes})
        if res["OK"]:
            idleTransformations = res["Value"]
            idleTransformations = [d.get("TransformationID") for d in idleTransformations]
            self.log.verbose(f"Found {len(idleTransformations)} Idle MC transformations")
            self.log.debug(f"Idle transformations found: {','.join([str(it) for it in idleTransformations])}")
        else:
            self.log.error("Call to Transformation Client service failed", res["Message"])
            return res

        # get all the IDs of transformations undergoing a testing phase
        res = self.transClient.getStoredJobDescriptionIDs()
        if res["OK"]:
            testingSimulations = res["Value"]
            testingSimulations = [pair[0] for pair in testingSimulations]
            self.log.verbose(f"Found {len(testingSimulations)} MC transformations undergoing a testing phase")
            self.log.debug(
                "MC transformations found undergoing a testing phase: %s"
                % ",".join([str(ts) for ts in testingSimulations])
            )
        else:
            self.log.error("Call to Transformation Client service failed", res["Message"])
            return res

        # get the IDs that occur in both idle transformations and testing phase
        idleSimulations = list(set(testingSimulations).intersection(idleTransformations))
        # remove those that we know failed
        idleSimulations = list(set(idleSimulations).difference(self.failedTransIDs))
        self.log.info(
            "MC transformations under considerations: %s (will loop on them)"
            % ",".join([str(idS) for idS in idleSimulations])
        )
        for transID in idleSimulations:
            self.log.info("Looking into %d" % transID)
            tasks = self.transClient.getTransformationTasks(condDict={"TransformationID": transID})
            if not tasks["OK"]:
                self.log.error("Call to Transformation Client service failed", tasks["Message"])
                continue
            else:
                tasks = tasks["Value"]
                numberOfTasks = len(tasks)
                numberOfDoneTasks = sum(1 for d in tasks if d.get("ExternalStatus") == "Done")
                self.log.verbose(
                    "TransID = %d, numberOfTasks = %d, numberOfDoneTasks = %d"
                    % (transID, numberOfTasks, numberOfDoneTasks)
                )
                if numberOfTasks == numberOfDoneTasks:
                    self.log.info("All tasks have passed so the request can be accepted and the transformation updated")
                    res = self._activateTransformation(transID, tasks)
                    if not res["OK"]:
                        self.log.error("Error Activating Production", res["Message"])
                else:
                    self.log.warn("There are failed tasks")
                    report = self.__createReport(tasks)
                    numberOfFailedTasks = sum(1 for d in tasks if d.get("ExternalStatus") == "Failed")
                    if numberOfFailedTasks == numberOfTasks:
                        # all tasks have failed so the request can be rejected and an email report sent
                        self._sendReport(report)
                        self.log.warn("Transformation " + str(transID) + " failed the testing phase")
                        self.failedTransIDs.append(transID)
                    else:
                        # only some tasks have failed so continue but send a warn email
                        self.log.warn(
                            "Transformation " + str(transID) + " failed partially the testing phase, continuing anyway"
                        )
                        doneTasks = list()
                        for d in tasks:
                            if d.get("ExternalStatus") == "Done":
                                doneTasks.append(d)
                        if not doneTasks:
                            self.log.info("No tasks done for Transformation %d" % transID)
                            continue
                        res = self._activateTransformation(transID, doneTasks)
                        if not res["OK"]:
                            self.log.error("Error Activating Production", res["Message"])
                            continue
                        subject = (
                            "MCSimulation Test Failure Report. TransformationID: "
                            + str(transID)
                            + " - some tasks failed"
                        )
                        report["subject"] = subject
                        self._sendReport(report)

        return S_OK()

    def _activateTransformation(self, transID, tasks):
        """Calculate parameters, update the workflow, then move the production to
        Active."""
        parameters = self._calculateParameters(tasks)
        if not parameters["OK"]:
            self.log.error("Error calculating parameters", parameters["Message"])
            return parameters
        else:
            parameters = parameters["Value"]
            self.log.verbose("TransID = %d, Calculated Parameters: %s" % (transID, str(parameters)))
            workflow = self._updateWorkflow(transID, int(round(float(parameters["CPUe"]))), parameters["MCCpu"])
            if workflow["OK"]:
                workflow = workflow["Value"]
                res = self._updateTransformationsTable(transID, workflow)
                if not res["OK"]:
                    self.log.error("Error updating transformations table", res["Message"])
                    return res
                else:
                    self.log.info(
                        "Transformation " + str(transID) + " passed the testing phase and is now set to active"
                    )

        return S_OK()

    def __createReport(self, tasks):
        """creates a report from a failed task to email to the production
        manager."""
        dateformat = "%d/%m/%Y %H:%M"
        transformationID = tasks[0]["TransformationID"]
        transformation = self.transClient.getTransformations(condDict={"TransformationID": transformationID})
        transformation = transformation["Value"][0]
        subject = "MCSimulation Test Failure Report. TransformationID: " + str(transformationID)
        body = [subject]
        body.append("")
        body.append("Transformation:")
        body.append("----------------------------------------------------------------------")
        body.append("TransformationID: " + str(transformation["TransformationID"]))
        body.append("TransformationName: " + transformation["TransformationName"])
        body.append("LastUpdate: " + transformation["LastUpdate"].strftime(dateformat))
        body.append("Status: " + transformation["Status"])
        body.append("Description: " + transformation["Description"])
        body.append("TransformationFamily: " + str(transformation["TransformationFamily"]))
        body.append("Plugin: " + transformation["Plugin"])
        body.append("Type: " + transformation["Type"])
        body.append("AgentType: " + transformation["AgentType"])
        body.append("GroupSize: " + str(transformation["GroupSize"]))
        body.append("MaxNumberOfTasks: " + str(transformation["MaxNumberOfTasks"]))
        body.append("AuthorDN: " + transformation["AuthorDN"])
        body.append("TransformationGroup: " + transformation["TransformationGroup"])
        body.append("InheritedFrom: " + str(transformation["InheritedFrom"]))
        body.append("CreationDate: " + transformation["CreationDate"].strftime(dateformat))
        body.append("FileMask: " + transformation["FileMask"])
        body.append("EventsPerTask: " + str(transformation["EventsPerTask"]))
        body.append("AuthorGroup: " + transformation["AuthorGroup"])
        body.append("")
        body.append("Number of Tasks: " + str(len(tasks)))
        body.append("Tasks:")
        body.append("----------------------------------------------------------------------")
        for task in tasks:
            body.append("TaskID: " + str(task["TaskID"]))
            body.append("TargetSE: " + task["TargetSE"])
            body.append("LastUpdateTime: " + task["LastUpdateTime"].strftime(dateformat))
            body.append("RunNumber: " + str(task.get("RunNumber", 0)))
            body.append("CreationTime: " + task["CreationTime"].strftime(dateformat))
            body.append("ExternalID: " + str(task["ExternalID"]))
            body.append("ExternalStatus: " + task["ExternalStatus"])
            body.append("")
        return {"subject": subject, "body": body}

    def _sendReport(self, report):
        """sends a given report to the production manager."""
        if not self.email:
            self.email = getUserOption(self.operations.getValue("Shifter/ProductionManager/User"), "Email")
        body = "\n".join(report["body"])
        res = self.notifyClient.sendMail(self.email, report["subject"], body, self.email, localAttempt=False)
        if not res["OK"]:
            self.log.error("sendMail failed", res["Message"])
        else:
            self.log.info("Mail summary sent to production manager")

    def _calculateParameters(self, tasks):
        """Calculates the CPU time per event for the production."""
        jobIds = [int(x["ExternalID"]) for x in tasks]
        res = self.bkClient.bulkJobInfo({"jobId": jobIds})
        if not res["OK"]:
            self.log.error("Error calling bkClient", res["Message"])
            return S_ERROR(res["Message"])
        successful = res["Value"]["Successful"]
        self.log.debug(f"Successful tasks: {str(successful)}")
        if not successful:
            self.log.error("There are no successful tasks")
            return S_ERROR("There are no successful tasks")

        successfulJobs = len(successful)

        CPUeJobTotal = 0.0
        for job in successful.values():
            cpuJob = 0
            events = None
            for bkJob in job:
                if bkJob["ApplicationName"] in ["Gauss", "Boole", "Moore", "Brunel", "DaVinci"]:
                    if bkJob["ApplicationName"] == "Gauss":
                        events = bkJob["NumberOfEvents"]  # This must be there
                    timeInSeconds = bkJob["CPUTIME"]
                    cpuJob += timeInSeconds * bkJob["WNCPUHS06"]
            if not events:  # extreme case of no events produced
                successfulJobs -= 1
                continue
            CPUeJob = cpuJob / events
            self.log.debug("CPUeJob = %d" % CPUeJob)

            CPUeJobTotal += CPUeJob

        CPUe = CPUeJobTotal / successfulJobs
        # We want to produce at least 25 events per job...
        MCCpu = str(25 * int(round(float(CPUe))))
        self.log.verbose("CPUe = %d, MCCpu = %s" % (CPUe, MCCpu))
        return S_OK({"CPUe": CPUe, "MCCpu": MCCpu})

    def _updateWorkflow(self, transID, CPUe, MCCpu):
        """Updates the workflow of a savedProductionDescription to reflect the
        calculated CPUe."""
        res = self.transClient.getStoredJobDescription(transID)
        if res["OK"]:
            workflow = fromXMLString(res["Value"][0][1])
            prod = Production()
            prod.LHCbJob.workflow = workflow
            prod.setParameter("CPUe", "string", str(CPUe), "CPU time per event")
            prod.LHCbJob.setCPUTime(MCCpu)
            self.log.info("Transformation ", str(transID))
            self.log.info("Calculated CPUTime: ", str(CPUe))
            self.log.info("CpuTime: ", str(MCCpu))

            # maximum number of events to produce
            # try to get the CPU parameters from the configuration if possible
            cpuTimeAvg = Operations().getValue("Transformations/CPUTimeAvg")
            if cpuTimeAvg is None:
                self.log.info("Could not get CPUTimeAvg from config, defaulting to %d" % 200000)
                cpuTimeAvg = 200000

            try:
                CPUNormalizationFactorAvg = getCPUNormalizationFactorAvg()
            except RuntimeError:
                self.log.info("Could not get CPUNormalizationFactorAvg, defaulting to %f" % 1.0)
                CPUNormalizationFactorAvg = 1.0

            max_e = getEventsToProduce(CPUe, cpuTimeAvg, CPUNormalizationFactorAvg)
            prod.setParameter("maxNumberOfEvents", "string", str(max_e), "Maximum number of events to produce (Gauss)")
            return S_OK(prod.LHCbJob.workflow.toXML())
        else:
            self.log.error("Call to Transformation Client service failed", res["Message"])
            return res

    def _updateTransformationsTable(self, transID, workflow):
        """Puts the modified workflow from the savedProductionDescription table
        into the transformations table and removes it from the
        savedProductionDescription table."""
        transformation = self.transClient.getTransformations(condDict={"TransformationID": transID})
        if transformation["OK"]:
            body = self.transClient.setTransformationParameter(transID, "Body", workflow)
            status = self.transClient.setTransformationParameter(transID, "Status", "Active")
            if body["OK"] and status["OK"]:
                res = self.transClient.removeStoredJobDescription(transID)
                if not res["OK"]:
                    self.log.error("Call to removeStoredJobDescription failed", res["Message"])
                    return res
                self.log.info(f"Transformation {transID} has an updated body and Status set to active")
                return S_OK()
            else:
                self.log.error(
                    "One of the updates has failed so set them both back to the previous value to ensure atomicity"
                )
                self.log.debug(str(transformation["Value"][0]["Body"]))
                res = self.transClient.setTransformationParameter(transID, "Body", transformation["Value"][0]["Body"])
                if not res["OK"]:
                    self.log.error("Failure calling setTransformationParameter", res["Message"])
                    return res
                res = self.transClient.setTransformationParameter(
                    transID, "Status", transformation["Value"][0]["Status"]
                )
                if not res["OK"]:
                    self.log.error("Failure calling setTransformationParameter", res["Message"])
                    return res
        else:
            self.log.error("Call to getTransformations failed", transformation["Message"])
            return transformation



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCb Bookkeeping database client."""
import os
from LHCbDIRAC.BookkeepingSystem.Client.LHCbBookkeepingManager import LHCbBookkeepingManager


class LHCB_BKKDBClient:
    """Client used to browse the Entities."""

    def __init__(self):
        """Initialize the basic class."""

        self.__ESManager = LHCbBookkeepingManager()
        result = self.__ESManager.getAbsolutePath("/")
        if result["OK"]:
            self.__currentDirectory = result["Value"]

    def help(self):
        """help function."""
        return self.__ESManager.help()  # pylint: disable=no-member

    def list(self, path="", selectionDict=None, sortDict=None, startItem=0, maxitems=0):
        """It lists the database content as a Linux File System."""
        selectionDict = selectionDict if selectionDict is not None else {}
        sortDict = sortDict if sortDict is not None else {}
        res = self.__ESManager.getAbsolutePath(os.path.join(self.__currentDirectory, path))
        if not res["OK"]:
            return res
        return self.__ESManager.list(res["Value"], selectionDict, sortDict, startItem, maxitems)

    def get(self, path=""):
        """get path."""
        return self.__ESManager.get(path)

    def getPossibleParameters(self):
        """available trees."""
        return self.__ESManager.getPossibleParameters()  # pylint: disable=no-member

    def setParameter(self, name):
        """tree used."""
        return self.__ESManager.setParameter(name)  # pylint: disable=no-member

    def getLogicalFiles(self):
        """lfns."""
        return self.__ESManager.files_  # pylint: disable=no-member

    def getFilesPFN(self):
        """PFNS."""
        return self.__ESManager.getFilesPFN()  # pylint: disable=no-member

    def getNumberOfEvents(self, files):
        """number of events."""
        return self.__ESManager.getNumberOfEvents(files)  # pylint: disable=no-member

    def writeJobOptions(
        self, files, optionsFile="jobOptions.opts", savedType=None, catalog=None, savePfn=None, dataset=None
    ):
        """Gaudi card."""
        return self.__ESManager.writeJobOptions(  # pylint: disable=no-member
            files, optionsFile, savedType, catalog, savePfn, dataset
        )

    def getJobInfo(self, lfn):
        """how a file is created."""
        return self.__ESManager.getJobInfo(lfn)  # pylint: disable=no-member

    def setAdvancedQueries(self, value):
        """Advanced queries."""
        return self.__ESManager.setAdvancedQueries(value)  # pylint: disable=no-member

    def getLimitedFiles(self, selectionDict, sortDict, startItem, maxitems):
        """get files used by Web portal."""
        return self.__ESManager.getLimitedFiles(  # pylint: disable=no-member
            selectionDict, sortDict, startItem, maxitems
        )

    def getAncestors(self, files, depth):
        """ancestor of files."""
        return self.__ESManager.getAncestors(files, depth)  # pylint: disable=no-member

    def getFileCreationLog(self, filename):
        """log file of a given file."""
        return self.__ESManager.getFileCreationLog(filename)  # pylint: disable=no-member

    def writePythonOrJobOptions(self, startItem, maxitems, path, optstype):
        """python job option."""
        return self.__ESManager.writePythonOrJobOptions(  # pylint: disable=no-member
            startItem, maxitems, path, optstype
        )

    def getLimitedInformations(self, startItem, maxitems, path):
        """statistics."""
        return self.__ESManager.getLimitedInformations(startItem, maxitems, path)  # pylint: disable=no-member

    def getProcessingPassSteps(self, in_dict):
        """step."""
        return self.__ESManager.getProcessingPassSteps(in_dict)  # pylint: disable=no-member

    def getMoreProductionInformations(self, prodid):
        """production details."""
        return self.__ESManager.getMoreProductionInformations(prodid)  # pylint: disable=no-member

    def getAvailableProductions(self):
        """available productions."""
        return self.__ESManager.getAvailableProductions()  # pylint: disable=no-member

    def getFileHistory(self, lfn):
        """ "file history."""
        return self.__ESManager.getFileHistory(lfn)  # pylint: disable=no-member

    def getCurrentParameter(self):
        """curent bookkeeping path."""
        return self.__ESManager.getCurrentParameter()  # pylint: disable=no-member

    def getQueriesTypes(self):
        """type of the current query."""
        return self.__ESManager.getQueriesTypes()  # pylint: disable=no-member

    def getProductionProcessingPassSteps(self, in_dict):
        """the steps which produced a given production."""
        return self.__ESManager.getProductionProcessingPassSteps(in_dict)  # pylint: disable=no-member

    def getAvailableDataQuality(self):
        """available data quality."""
        return self.__ESManager.getAvailableDataQuality()  # pylint: disable=no-member

    def getAvailableExtendedDQOK(self):
        """available extended DQOK systems."""
        return self.__ESManager.getAvailableExtendedDQOK()  # pylint: disable=no-member

    def getAvailableSMOG2States(self):
        """available SMOG2 states."""
        return self.__ESManager.getAvailableSMOG2States()  # pylint: disable=no-member

    def setDataQualities(self, values):
        """set data qualities."""
        self.__ESManager.setDataQualities(values)  # pylint: disable=no-member

    def getStepsMetadata(self, bkDict):
        """returns detailed step metadata."""
        return self.__ESManager.getStepsMetadata(bkDict)  # pylint: disable=no-member

    def setFileTypes(self, fileTypeList):
        """it sets the file types."""
        return self.__ESManager.setFileTypes(fileTypeList)  # pylint: disable=no-member

    def getFilesWithMetadata(self, dataset):
        """it sets the file types.

        :param dict dataset: it is a bookkeeping dictionary, which contains the conditions used to retreive the lfns
        :return: S_OK lfns with metadata
        """
        return self.__ESManager.getFilesWithMetadata(dataset)  # pylint: disable=no-member



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for Workflow Modules utilities."""
# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

import unittest
import os

from unittest.mock import patch

from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers import Resources

from LHCbDIRAC.Workflow.Modules.ModulesUtilities import (
    lowerExtension,
    getEventsToProduce,
    getCPUNormalizationFactorAvg,
    getProductionParameterValue,
)


class ModulesUtilitiesTestCase(unittest.TestCase):
    """Base class for the ModulesUtilities test cases."""

    def setUp(self):
        gLogger.setLevel("DEBUG")
        self.maxDiff = None

    def tearDown(self):
        for fileProd in ["foo.txt", "BAR.txt", "FooBAR.ext.txt"]:
            try:
                os.remove(fileProd)
            except OSError:
                continue


#############################################################################
# ModulesUtilities.py
#############################################################################


class ModulesUtilitiesSuccess(ModulesUtilitiesTestCase):
    #################################################

    def test_lowerExtension(self):
        open("foo.tXt", "w").close()
        open("BAR.txt", "w").close()
        open("FooBAR.eXT.TXT", "w").close()

        lowerExtension()

        self.assertTrue("foo.txt" in os.listdir("."))
        self.assertTrue("BAR.txt" in os.listdir("."))
        self.assertTrue("FooBAR.ext.txt" in os.listdir("."))

    #################################################

    def test_getEventsToProduce(self):
        CPUe = 2.0
        CPUTime = 1000000.0
        CPUNormalizationFactor = 0.5

        out = getEventsToProduce(CPUe, CPUTime, CPUNormalizationFactor)
        outExp = 187500
        self.assertEqual(out, outExp)

        out = getEventsToProduce(CPUe, CPUTime, CPUNormalizationFactor, maxNumberOfEvents=1000)
        outExp = 1000
        self.assertEqual(out, outExp)

        out = getEventsToProduce(CPUe, CPUTime, CPUNormalizationFactor, jobMaxCPUTime=100000)
        outExp = 18750
        self.assertEqual(out, outExp)

        out = getEventsToProduce(CPUe, CPUTime, CPUNormalizationFactor, maxNumberOfEvents=1000, jobMaxCPUTime=100000)
        outExp = 1000
        self.assertEqual(out, outExp)

    #################################################

    def test_getCPUNormalizationFactorAvg(self):
        with patch.object(gConfig, "getSections") as mockGetSections:  # @UndefinedVariable
            with patch.object(Resources, "getQueues") as mockGetQueues:  # @UndefinedVariable
                # gConfig.getSection error
                mockGetSections.return_value = S_ERROR()
                self.assertRaises(RuntimeError, getCPUNormalizationFactorAvg)

                # Resources.getQueues error
                mockGetSections.return_value = S_OK(["LCG.CERN.ch"])
                mockGetQueues.return_value = S_ERROR()
                self.assertRaises(RuntimeError, getCPUNormalizationFactorAvg)

                # no queues
                mockGetQueues.return_value = S_OK({"LCG.CERN.ch": {}})
                self.assertRaises(RuntimeError, getCPUNormalizationFactorAvg)

                # success
                mockGetQueues.return_value = S_OK(
                    {
                        "LCG.CERN.ch": {
                            "ce201.cern.ch": {
                                "CEType": "CREAM",
                                "OS": "ScientificCERNSLC_Boron_5.5",
                                "Pilot": "True",
                                "Queues": {
                                    "cream-lsf-grid_2nh_lhcb": {
                                        "MaxTotalJobs": "1000",
                                        "MaxWaitingJobs": "20",
                                        "SI00": "1000",
                                        "maxCPUTime": "120",
                                    },
                                    "cream-lsf-grid_lhcb": {
                                        "MaxTotalJobs": "1000",
                                        "MaxWaitingJobs": "100",
                                        "SI00": "1000",
                                        "WaitingToRunningRatio": "0.2",
                                        "maxCPUTime": "10080",
                                    },
                                },
                                "SI00": "5242",
                                "SubmissionMode": "Direct",
                                "architecture": "x86_64",
                                "wnTmpDir": ".",
                            },
                            "ce202.cern.ch": {
                                "CEType": "CREAM",
                                "OS": "ScientificCERNSLC_Boron_5.8",
                                "Pilot": "True",
                                "Queues": {
                                    "cream-lsf-grid_2nh_lhcb": {
                                        "MaxTotalJobs": "1000",
                                        "MaxWaitingJobs": "20",
                                        "SI00": "1000",
                                        "maxCPUTime": "120",
                                    },
                                    "cream-lsf-grid_lhcb": {
                                        "MaxTotalJobs": "1000",
                                        "MaxWaitingJobs": "100",
                                        "SI00": "1000",
                                        "WaitingToRunningRatio": "0.2",
                                        "maxCPUTime": "10080",
                                    },
                                },
                                "SI00": "5242",
                                "SubmissionMode": "Direct",
                                "architecture": "x86_64",
                                "wnTmpDir": ".",
                            },
                        }
                    }
                )
                out = getCPUNormalizationFactorAvg()
                self.assertEqual(out, 4.0)

    #################################################

    def test_getProductionParameterValue(self):
        emptyXML = "<Workflow></Workflow>"
        noValueProductionXML = """
      <Workflow>
        <origin></origin>
        <description><![CDATA[prodDescription]]></description>
        <descr_short></descr_short>
        <version>0.0</version>
        <type></type>
        <name>Request_12416_MCSimulation_Sim08a/Digi13/Trig0x40760037/Reco14a/Stripping20r1NoPrescalingFlagged_EventType_13296003__1</name>
        <Parameter name="JobType" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="User specified type">
          <value></value>
        </Parameter>
      </Workflow>
    """  # noqa
        productionXML = """
      <Workflow>
        <origin></origin>
        <description><![CDATA[prodDescription]]></description>
        <descr_short></descr_short>
        <version>0.0</version>
        <type></type>
        <name>Request_12416_MCSimulation_Sim08a/Digi13/Trig0x40760037/Reco14a/Stripping20r1NoPrescalingFlagged_EventType_13296003__1</name>
        <Parameter name="JobType" type="JDL" linked_module="" linked_parameter="" in="True" out="False" description="User specified type">
          <value><![CDATA[MCSimulation]]></value>
        </Parameter>
      </Workflow>
    """  # noqa

        parameterName = "JobType"

        valueExp = "MCSimulation"

        value = getProductionParameterValue(emptyXML, parameterName)
        self.assertEqual(value, None)

        value = getProductionParameterValue(noValueProductionXML, parameterName)
        self.assertEqual(value, None)

        value = getProductionParameterValue(productionXML, parameterName)
        self.assertEqual(value, valueExp)


#############################################################################
# Test Suite run
#############################################################################

if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(ModulesUtilitiesTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ModulesUtilitiesSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: StorageUsageAgent

.. module: StorageUsageAgent

:synopsis: StorageUsageAgent takes the FC as the primary source of information to
  determine storage usage.
"""
# # imports
import datetime
import time
import random
import os
import re
import threading

# # from DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.Core.Utilities.DirectoryExplorer import DirectoryExplorer
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Core.Utilities import List
from DIRAC.Core.Utilities.TimeUtilities import timeInterval, week
from DIRAC.Core.Utilities.DictCache import DictCache
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

# # from LHCbDIRAC
from LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB import StorageUsageDB
from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient

AGENT_NAME = "DataManagement/StorageUsageAgent"


def removeProxy(proxyFile):
    if os.path.exists(proxyFile):
        os.remove(proxyFile)


class StorageUsageAgent(AgentModule):
    """.. class:: StorageUsageAgent.

    :param FileCatalog catalog: FileCatalog instance
    :parma mixed storageUsage: StorageUsageDB instance or its rpc client
    :param int pollingTime: polling time
    :param int activePeriod: active period on weeks
    :param threading.Lock dataLock: data lock
    :param threading.Lock replicaListLock: replica list lock
    :param DictCache proxyCache: creds cache
    """

    catalog = None
    storageUsage = None
    pollingTime = 43200
    activePeriod = 0
    dataLock = None  # threading.Lock()
    replicaListLock = None  # threading.Lock()
    proxyCache = None  # DictCache()
    enableStartupSleep = True  # Enable a random sleep so not all the user agents start together

    def __init__(self, *args, **kwargs):
        """c'tor."""
        AgentModule.__init__(self, *args, **kwargs)

        self.__baseDir = "/lhcb"
        self.__baseDirLabel = "_".join(List.fromChar(self.__baseDir, "/"))
        self.__ignoreDirsList = []
        self.__keepDirLevels = 4

        self.__startExecutionTime = int(time.time())
        self.__dirExplorer = DirectoryExplorer(reverse=True)
        self.__processedDirs = 0
        self.__directoryOwners = {}
        self.catalog = FileCatalog()
        self.__maxToPublish = self.am_getOption("MaxDirectories", 5000)
        if self.am_getOption("DirectDB", False):
            self.storageUsage = StorageUsageDB()
        else:
            # Set a timeout of 0.1 seconds per directory (factor 5 margin)
            self.storageUsage = StorageUsageClient(timeout=self.am_getOption("Timeout", int(self.__maxToPublish * 0.1)))
        self.activePeriod = self.am_getOption("ActivePeriod", self.activePeriod)
        self.dataLock = threading.Lock()
        self.replicaListLock = threading.Lock()
        self.proxyCache = DictCache(removeProxy)
        self.__noProxy = set()
        self.__catalogType = None
        self.__recalculateUsage = Operations().getValue("DataManagement/RecalculateDirSize", False)
        self.enableStartupSleep = self.am_getOption("EnableStartupSleep", self.enableStartupSleep)
        self.__publishDirQueue = {}
        self.__dirsToPublish = {}
        self.__replicaFilesUsed = set()
        self.__replicaListFilesDir = ""

    def initialize(self):
        """agent initialisation."""

        self.am_setOption("PollingTime", self.pollingTime)

        if self.enableStartupSleep:
            rndSleep = random.randint(1, self.pollingTime)
            self.log.info(f"Sleeping for {rndSleep} seconds")
            time.sleep(rndSleep)

        # This sets the Default Proxy to used as that defined under
        # /Operations/Shifter/DataManager
        # the shifterProxy option in the Configsorteduration can be used to change this default.
        self.am_setOption("shifterProxy", "DataManager")

        return S_OK()

    def __writeReplicasListFiles(self, dirPathList):
        """dump replicas list to files."""
        self.replicaListLock.acquire()
        try:
            self.log.info(f"Dumping replicas for {len(dirPathList)} dirs")
            result = self.catalog.getDirectoryReplicas(dirPathList)
            if not result["OK"]:
                self.log.error("Could not get directory replicas", f"{dirPathList} -> {result['Message']}")
                return result
            resData = result["Value"]
            filesOpened = {}
            for dirPath in dirPathList:
                if dirPath in result["Value"]["Failed"]:
                    self.log.error("Could not get directory replicas", f"{dirPath} -> {resData['Failed'][dirPath]}")
                    continue
                dirData = resData["Successful"][dirPath]
                for lfn in dirData:
                    for seName in dirData[lfn]:
                        if seName not in filesOpened:
                            filePath = os.path.join(
                                self.__replicaListFilesDir, f"replicas.{seName}.{self.__baseDirLabel}.filling"
                            )
                            # Check if file is opened and if not open it
                            if seName not in filesOpened:
                                if seName not in self.__replicaFilesUsed:
                                    self.__replicaFilesUsed.add(seName)
                                    filesOpened[seName] = open(filePath, "w")
                                else:
                                    filesOpened[seName] = open(filePath, "a")
                        # seName file is opened. Write
                        filesOpened[seName].write(f"{lfn} -> {dirData[lfn][seName]}\n")
            # Close the files
            for seName in filesOpened:
                filesOpened[seName].close()
            return S_OK()
        finally:
            self.replicaListLock.release()

    def __resetReplicaListFiles(self):
        """prepare directories for replica list files."""
        self.__replicaFilesUsed = set()
        self.__replicaListFilesDir = os.path.join(self.am_getOption("WorkDirectory"), "replicaLists")
        mkDir(self.__replicaListFilesDir)
        self.log.info(f"Replica Lists directory is {self.__replicaListFilesDir}")

    def __replicaListFilesDone(self):
        """rotate replicas list files."""
        self.replicaListLock.acquire()
        try:
            old = re.compile(r"^replicas\.([a-zA-Z0-9\-_]*)\.%s\.old$" % self.__baseDirLabel)
            current = re.compile(r"^replicas\.([a-zA-Z0-9\-_]*)\.%s$" % self.__baseDirLabel)
            filling = re.compile(r"^replicas\.([a-zA-Z0-9\-_]*)\.%s\.filling$" % self.__baseDirLabel)
            # Delete old
            for fileName in os.listdir(self.__replicaListFilesDir):
                match = old.match(fileName)
                if match:
                    os.unlink(os.path.join(self.__replicaListFilesDir, fileName))
            # Current -> old
            for fileName in os.listdir(self.__replicaListFilesDir):
                match = current.match(fileName)
                if match:
                    newFileName = f"replicas.{match.group(1)}.{self.__baseDirLabel}.old"
                    self.log.info(
                        "Moving \n %s\n to \n %s"
                        % (
                            os.path.join(self.__replicaListFilesDir, fileName),
                            os.path.join(self.__replicaListFilesDir, newFileName),
                        )
                    )
                    os.rename(
                        os.path.join(self.__replicaListFilesDir, fileName),
                        os.path.join(self.__replicaListFilesDir, newFileName),
                    )
            # filling to current
            for fileName in os.listdir(self.__replicaListFilesDir):
                match = filling.match(fileName)
                if match:
                    newFileName = f"replicas.{match.group(1)}.{self.__baseDirLabel}"
                    self.log.info(
                        "Moving \n %s\n to \n %s"
                        % (
                            os.path.join(self.__replicaListFilesDir, fileName),
                            os.path.join(self.__replicaListFilesDir, newFileName),
                        )
                    )
                    os.rename(
                        os.path.join(self.__replicaListFilesDir, fileName),
                        os.path.join(self.__replicaListFilesDir, newFileName),
                    )

            return S_OK()
        finally:
            self.replicaListLock.release()

    def __printSummary(self):
        """pretty print summary."""
        res = self.storageUsage.getStorageSummary()
        if res["OK"]:
            self.log.notice("Storage Usage Summary")
            self.log.notice("============================================================")
            self.log.notice("%-40s %20s %20s" % ("Storage Element", "Number of files", "Total size"))

            time.sleep(2)

            for se in sorted(res["Value"]):
                usage = res["Value"][se]["Size"]
                files = res["Value"][se]["Files"]
                self.log.notice("%-40s %20s %20s" % (se, str(files), str(usage)))

    def execute(self):
        """execution in one cycle."""
        self.__publishDirQueue = {}
        self.__dirsToPublish = {}
        self.__baseDir = self.am_getOption("BaseDirectory", "/lhcb")
        self.__baseDirLabel = "_".join(List.fromChar(self.__baseDir, "/"))
        self.__ignoreDirsList = self.am_getOption("Ignore", [])
        self.__keepDirLevels = self.am_getOption("KeepDirLevels", 4)

        self.__startExecutionTime = int(time.time())
        self.__dirExplorer = DirectoryExplorer(reverse=True)
        self.__resetReplicaListFiles()
        self.__noProxy = set()
        self.__processedDirs = 0
        self.__directoryOwners = {}

        self.__printSummary()

        self.__dirExplorer.addDir(self.__baseDir)
        self.log.notice(f"Initiating with {self.__baseDir} as base directory.")
        # Loop over all the directories and sub-directories
        totalIterTime = 0.0
        numIterations = 0.0
        iterMaxDirs = 100
        while self.__dirExplorer.isActive():
            startT = time.time()
            d2E = [self.__dirExplorer.getNextDir() for _i in range(iterMaxDirs) if self.__dirExplorer.isActive()]
            self.__exploreDirList(d2E)
            iterTime = time.time() - startT
            totalIterTime += iterTime
            numIterations += len(d2E)
            self.log.verbose(f"Query took {iterTime:.2f} seconds for {len(d2E)} dirs")
        self.log.verbose(f"Average query time: {totalIterTime / numIterations:.2f} secs/dir")

        # Publish remaining directories
        self.__publishData(background=False)

        # Move replica list files
        self.__replicaListFilesDone()

        # Clean records older than 1 day
        self.log.info("Finished recursive directory search.")

        if self.am_getOption("PurgeOutdatedRecords", True):
            elapsedTime = time.time() - self.__startExecutionTime
            outdatedSeconds = max(max(self.am_getOption("PollingTime"), elapsedTime) * 2, 86400)
            result = self.storageUsage.purgeOutdatedEntries(self.__baseDir, int(outdatedSeconds), self.__ignoreDirsList)
            if not result["OK"]:
                return result
            self.log.notice(f"Purged {result['Value']} outdated records")
        return S_OK()

    def __exploreDirList(self, dirList):
        """collect directory size for directory in :dirList:"""
        # Normalise dirList first
        dirList = [os.path.realpath(d) for d in dirList]
        self.log.notice(f"Retrieving info for {len(dirList)} dirs")
        # For top directories, no files anyway, hence no need to get full size
        dirContents = {}
        failed = {}
        successfull = {}
        startTime = time.time()
        nbDirs = len(dirList)
        chunkSize = 10
        if self.__catalogType == "DFC" or dirList == [self.__baseDir]:
            # Get the content of the directory as anyway this is needed
            for dirChunk in breakListIntoChunks(dirList, chunkSize):
                res = self.catalog.listDirectory(dirChunk, True, timeout=600)
                if not res["OK"]:
                    failed.update(dict.fromkeys(dirChunk, res["Message"]))
                else:
                    failed.update(res["Value"]["Failed"])
                    dirContents.update(res["Value"]["Successful"])
            self.log.info(
                "Time to retrieve content of %d directories: %.1f seconds" % (nbDirs, time.time() - startTime)
            )
            for dirPath in failed:
                dirList.remove(dirPath)
        # We don't need to get the storage usage if there are no files...
        dirListSize = [d for d in dirList if dirContents.get(d, {}).get("Files")]

        startTime1 = time.time()
        # __recalculateUsage enables to recompute the directory usage in case the internal table is wrong
        for args in [(d, True, self.__recalculateUsage) for d in breakListIntoChunks(dirListSize, chunkSize)]:
            res = self.catalog.getDirectorySize(*args, timeout=600)
            if not res["OK"]:
                failed.update(dict.fromkeys(args[0], res["Message"]))
            else:
                failed.update(res["Value"]["Failed"])
                successfull.update(res["Value"]["Successful"])
        errorReason = {}
        for dirPath in failed:
            error = str(failed[dirPath])
            errorReason.setdefault(error, []).append(dirPath)
        for error in errorReason:
            self.log.error(
                "Failed to get directory info", "- {} for:\n\t{}".format(error, "\n\t".join(errorReason[error]))
            )
        self.log.info(
            f"Time to retrieve size of {len(dirListSize)} directories: {time.time() - startTime1:.1f} seconds"
        )
        for dirPath in [d for d in dirList if d not in failed]:
            metadata = successfull.get(dirPath, {})
            if "SubDirs" in metadata:
                self.__processDir(dirPath, metadata)
            else:
                if not self.__catalogType:
                    self.log.info("Catalog type determined to be DFC")
                    self.__catalogType = "DFC"
                self.__processDirDFC(dirPath, metadata, dirContents[dirPath])
        self.log.info("Time to process %d directories: %.1f seconds" % (nbDirs, time.time() - startTime))
        notCommited = len(self.__publishDirQueue) + len(self.__dirsToPublish)
        self.log.notice(
            "%d dirs to be explored, %d done. %d not yet committed."
            % (self.__dirExplorer.getNumRemainingDirs(), self.__processedDirs, notCommited)
        )

    def __processDirDFC(self, dirPath, metadata, subDirectories):
        """gets the list of subdirs that the DFC doesn't return, set the metadata
        like the FC and then call the same method as for the FC.
        """
        if "SubDirs" not in subDirectories:
            self.log.error("No subdirectory item for directory", dirPath)
            return
        dirMetadata = {"Files": 0, "TotalSize": 0, "ClosedDirs": [], "SiteUsage": {}}
        if "PhysicalSize" in metadata:
            dirMetadata["Files"] = metadata["LogicalFiles"]
            dirMetadata["TotalSize"] = metadata["LogicalSize"]
            dirMetadata["SiteUsage"] = metadata["PhysicalSize"].copy()
            dirMetadata["SiteUsage"].pop("TotalFiles", None)
            dirMetadata["SiteUsage"].pop("TotalSize", None)
        subDirs = subDirectories["SubDirs"].copy()
        dirMetadata["SubDirs"] = subDirs
        dirUsage = dirMetadata["SiteUsage"]
        errorReason = {}
        for subDir in subDirs:
            self.__directoryOwners.setdefault(subDir, (subDirs[subDir]["Owner"], subDirs[subDir]["OwnerGroup"]))
            subDirs[subDir] = subDirs[subDir].get("CreationTime", datetime.datetime.utcnow())
            if dirUsage:
                # This part here is for removing the recursivity introduced by the DFC
                args = [subDir]
                if len(subDir.split("/")) > self.__keepDirLevels:
                    args += [True, self.__recalculateUsage]
                result = self.catalog.getDirectorySize(*args)
                if not result["OK"]:
                    errorReason.setdefault(str(result["Message"]), []).append(subDir)
                    continue
                else:
                    metadata = result["Value"]["Successful"].get(subDir)
                    if metadata:
                        dirMetadata["Files"] -= metadata["LogicalFiles"]
                        dirMetadata["TotalSize"] -= metadata["LogicalSize"]
                    else:
                        errorReason.setdefault(str(result["Value"]["Failed"][subDir], [])).append(subDir)
                        continue
                if "PhysicalSize" in metadata and dirUsage:
                    seUsage = metadata["PhysicalSize"]
                    seUsage.pop("TotalFiles", None)
                    seUsage.pop("TotalSize", None)
                    for se in seUsage:
                        if se not in dirUsage:
                            self.log.error("SE used in subdir but not in dir", se)
                        else:
                            dirUsage[se]["Files"] -= seUsage[se]["Files"]
                            dirUsage[se]["Size"] -= seUsage[se]["Size"]
        for error in errorReason:
            self.log.error(
                "Failed to get directory info", "- {} for:\n\t{}".format(error, "\n\t".join(errorReason[error]))
            )
        for se, usage in list(dirUsage.items()):
            # Both info should be 0 or #0
            if not usage["Files"] and not usage["Size"]:
                dirUsage.pop(se)
            elif not usage["Files"] * usage["Size"]:
                self.log.error("Directory inconsistent", f"{dirPath} @ {se}: {str(usage)}")
        return self.__processDir(dirPath, dirMetadata)

    def __processDir(self, dirPath, dirMetadata):
        """calculate nb of files and size of :dirPath:, remove it if it's empty."""
        subDirs = dirMetadata["SubDirs"]
        closedDirs = dirMetadata["ClosedDirs"]
        ##############################
        # FIXME: Until we understand while closed dirs are not working...
        ##############################
        closedDirs = []
        prStr = f"{dirPath}: found {len(subDirs) if subDirs else 'no'} sub-directories"
        if closedDirs:
            prStr += f", {len(closedDirs)} are closed (ignored)"
        for rmDir in closedDirs + self.__ignoreDirsList:
            subDirs.pop(rmDir, None)
        numberOfFiles = int(dirMetadata["Files"])
        totalSize = int(dirMetadata["TotalSize"])
        if numberOfFiles:
            prStr += f" and {numberOfFiles} files ({totalSize} bytes)"
        else:
            prStr += " and no files"
        self.log.notice(prStr)
        if closedDirs:
            self.log.verbose("Closed dirs:\n %s" % "\n".join(closedDirs))
        siteUsage = dirMetadata["SiteUsage"]
        if numberOfFiles > 0:
            dirData = {"Files": numberOfFiles, "TotalSize": totalSize, "SEUsage": siteUsage}
            self.__addDirToPublishQueue(dirPath, dirData)
            # Print statistics
            self.log.verbose("%-40s %20s %20s" % ("Storage Element", "Number of files", "Total size"))
            for storageElement in sorted(siteUsage):
                usageDict = siteUsage[storageElement]
                self.log.verbose("%-40s %20s %20s" % (storageElement, str(usageDict["Files"]), str(usageDict["Size"])))
        # If it's empty delete it
        elif len(subDirs) == 0 and len(closedDirs) == 0:
            if dirPath != self.__baseDir:
                self.removeEmptyDir(dirPath)
                return
        # We don't need the cached information about owner
        self.__directoryOwners.pop(dirPath, None)
        rightNow = datetime.datetime.utcnow()
        chosenDirs = [
            subDir
            for subDir in subDirs
            if not self.activePeriod or timeInterval(subDirs[subDir], self.activePeriod * week).includes(rightNow)
        ]

        self.__dirExplorer.addDirList(chosenDirs)
        self.__processedDirs += 1

    def __getOwnerProxy(self, dirPath):
        """get owner creds for :dirPath:"""
        self.log.verbose("Retrieving dir metadata...")
        # get owner form the cached information, if not, try getDirectoryMetadata
        ownerName, ownerGroup = self.__directoryOwners.pop(dirPath, (None, None))
        if not ownerName or not ownerGroup:
            result = returnSingleResult(self.catalog.getDirectoryMetadata(dirPath))
            if not result["OK"] or "OwnerRole" not in result["Value"]:
                self.log.error("Could not get metadata info", result["Message"])
                return result
            ownerRole = result["Value"]["OwnerRole"]
            ownerDN = result["Value"]["OwnerDN"]
            if ownerRole[0] != "/":
                ownerRole = f"/{ownerRole}"
            ownerName = "unknown"
            byGroup = False
        else:
            ownerDN = Registry.getDNForUsername(ownerName)
            if not ownerDN["OK"]:
                self.log.error("Could not get DN from user name", ownerDN["Message"])
                return ownerDN
            ownerDN = ownerDN["Value"][0]
            # This bloody method returns directly a string!!!!
            ownerRole = Registry.getVOMSAttributeForGroup(ownerGroup)
            byGroup = True
            # Get all groups for that VOMS Role, and add lhcb_user as in DFC this is a safe value
        ownerGroups = Registry.getGroupsWithVOMSAttribute(ownerRole) + ["lhcb_user"]

        downErrors = []
        for ownerGroup in ownerGroups:
            cacheKey = (ownerDN, ownerRole)
            if byGroup:
                ownerRole = None
                cacheKey = (ownerDN, ownerGroup)
            if cacheKey in self.__noProxy:
                return S_ERROR("Proxy not available")
                # Getting the proxy...
            upFile = self.proxyCache.get(cacheKey, 3600)
            if upFile and os.path.exists(upFile):
                self.log.verbose(
                    "Returning cached proxy for %s %s@%s [%s] in %s"
                    % (ownerName, ownerDN, ownerGroup, ownerRole, upFile)
                )
                return S_OK(upFile)
            if ownerRole:
                result = gProxyManager.downloadVOMSProxy(
                    ownerDN, ownerGroup, limited=False, requiredVOMSAttribute=ownerRole
                )
            else:
                result = gProxyManager.downloadProxy(ownerDN, ownerGroup, limited=False)
            if not result["OK"]:
                downErrors.append(f"{cacheKey} : {result['Message']}")
                continue
            userProxy = result["Value"]
            secsLeft = max(0, userProxy.getRemainingSecs()["Value"])
            upFile = userProxy.dumpAllToFile()
            if upFile["OK"]:
                upFile = upFile["Value"]
            else:
                return upFile
            self.proxyCache.add(cacheKey, secsLeft, upFile)
            self.log.info(f"Got proxy for {ownerName} {ownerDN}@{ownerGroup} [{ownerRole}]")
            return S_OK(upFile)
        self.__noProxy.add(cacheKey)
        return S_ERROR(
            "Could not download proxy for user ({}, {}):\n{} ".format(ownerDN, ownerRole, "\n ".join(downErrors))
        )

    def removeEmptyDir(self, dirPath):
        self.log.notice(f"Deleting empty directory {dirPath}")
        for useOwnerProxy in (False, True):
            result = self.__removeEmptyDir(dirPath, useOwnerProxy=useOwnerProxy)
            if result["OK"]:
                self.log.info("Successfully removed empty directory from File Catalog and StorageUsageDB")
                break
        return result

    def __removeEmptyDir(self, dirPath, useOwnerProxy=True):
        """unlink empty folder :dirPath:"""
        from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData

        if len(List.fromChar(dirPath, "/")) < self.__keepDirLevels:
            return S_OK()

        if useOwnerProxy:
            result = self.__getOwnerProxy(dirPath)
            if not result["OK"]:
                if "Proxy not available" not in result["Message"]:
                    self.log.error(result["Message"])
                return result

            upFile = result["Value"]
            prevProxyEnv = os.environ["X509_USER_PROXY"]
            os.environ["X509_USER_PROXY"] = upFile
        try:
            gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "false")
            # res = self.catalog.removeDirectory( dirPath )
            res = self.catalog.writeCatalogs[0][1].removeDirectory(dirPath)
            if not res["OK"]:
                self.log.error("Error removing empty directory from File Catalog.", res["Message"])
                return res
            elif dirPath in res["Value"]["Failed"]:
                self.log.error("Failed to remove empty directory from File Catalog.", res["Value"]["Failed"][dirPath])
                self.log.debug(str(res))
                return S_ERROR(res["Value"]["Failed"][dirPath])
            res = self.storageUsage.removeDirectory(dirPath)
            if not res["OK"]:
                self.log.error("Failed to remove empty directory from Storage Usage database.", res["Message"])
                return res
            return S_OK()
        finally:
            gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "true")
            if useOwnerProxy:
                os.environ["X509_USER_PROXY"] = prevProxyEnv

    def __addDirToPublishQueue(self, dirName, dirData):
        """enqueue :dirName: and :dirData: for publishing."""
        self.__publishDirQueue[dirName] = dirData
        numDirsToPublish = len(self.__publishDirQueue)
        if numDirsToPublish and numDirsToPublish % self.am_getOption("PublishClusterSize", 100) == 0:
            self.__publishData(background=True)

    def __publishData(self, background=True):
        """publish data in a separate deamon thread."""
        self.dataLock.acquire()
        try:
            # Dump to file
            if self.am_getOption("DumpReplicasToFile", False):
                pass
                # repThread = threading.Thread( target = self.__writeReplicasListFiles,
                #                              args = ( list( self.__publishDirQueue ), ) )
            self.__dirsToPublish.update(self.__publishDirQueue)
            self.__publishDirQueue = {}
        finally:
            self.dataLock.release()
        if background:
            pubThread = threading.Thread(target=self.__executePublishData)
            pubThread.setDaemon(1)
            pubThread.start()
        else:
            self.__executePublishData()

    def __executePublishData(self):
        """publication thread target."""
        self.dataLock.acquire()
        try:
            if not self.__dirsToPublish:
                self.log.info("No data to be published")
                return
            if len(self.__dirsToPublish) > self.__maxToPublish:
                toPublish = {}
                for dirName in sorted(self.__dirsToPublish)[: self.__maxToPublish]:
                    toPublish[dirName] = self.__dirsToPublish.pop(dirName)
            else:
                toPublish = self.__dirsToPublish
            self.log.info(f"Publishing usage for {len(toPublish)} directories")
            res = self.storageUsage.publishDirectories(toPublish)
            if res["OK"]:
                # All is OK, reset the dictionary, even if data member!
                toPublish.clear()
            else:
                # Put back dirs to be published, due to the error
                self.__dirsToPublish.update(toPublish)
                self.log.error("Failed to publish directories", res["Message"])
            return res
        finally:
            self.dataLock.release()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""stores the job configuration."""


class JobConfiguration:
    def __init__(self):
        """initialize the class members."""
        self.configName = ""  # None
        self.configVersion = ""  # None
        self.date = ""  # None
        self.time = ""  # None

    def __repr__(self):
        """formats the output of the print."""
        result = "JobConfiguration: \n"
        result += "ConfigName:" + self.configName + "\n"
        result += "ConfigVersion:" + self.configVersion + "\n"
        result += "Date and Time:" + self.date + " " + self.time
        return result

    def writeToXML(self):
        """creates an xml string."""
        result = (
            '<Job ConfigName="'
            + self.configName
            + '" ConfigVersion="'
            + self.configVersion
            + '" Date="'
            + self.date
            + '" Time="'
            + self.time
            + '">\n'
        )
        return result



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Client module to deal with transformations, but mostly dedicated to
DataManipulation (e.g.: replications)"""

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.TransformationSystem.Client.Transformation import Transformation as DIRACTransformation
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup

from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

COMPONENT_NAME = "Transformation"


class Transformation(DIRACTransformation):
    """Class for dealing with Transformation objects."""

    #############################################################################

    def __init__(self, transID=0, transClientIn=None):
        """Just params setting.

        transClient is passed here as LHCbDIRAC TransformationsClient, it
        will be self.transClient
        """

        if not transClientIn:
            self.transClient = TransformationClient()
        else:
            self.transClient = transClientIn

        super().__init__(transID=transID, transClient=self.transClient)

    #############################################################################

    def testBkQuery(self, bkQuery, printOutput=False, bkClient=None):
        """just pretty print of the result of a BK Query."""

        if bkClient is None:
            bkClient = BookkeepingClient()

        res = bkClient.getFiles(bkQuery)
        if not res["OK"]:
            return self._errorReport(res, "Failed to perform BK query")
        gLogger.info(f"The supplied query returned {len(res['Value'])} files")
        if printOutput:
            self._prettyPrint(res)
        return S_OK(res["Value"])

    #############################################################################

    def setBkQuery(self, queryDict, test=False):
        """set a BKK Query."""
        if test:
            res = self.testBkQuery(queryDict)
            if not res["OK"]:
                return res
        transID = self.paramValues["TransformationID"]
        if self.exists and transID:
            res = self.transClient.addBookkeepingQuery(transID, queryDict)
            if not res["OK"]:
                return res
        self.item_called = "BkQuery"
        self.paramValues[self.item_called] = queryDict
        return S_OK()

    #############################################################################

    def getBkQuery(self, printOutput=False):
        """get a BKK Query."""
        if self.paramValues["BkQuery"]:
            return S_OK(self.paramValues["BkQuery"])
        res = self.__executeOperation("getBookkeepingQuery")
        if not res["OK"]:
            return res
        self.item_called = "BkQuery"
        self.paramValues[self.item_called] = res["Value"]
        return S_OK(res["Value"])

    #############################################################################

    def deleteTransformationBkQuery(self):
        """delete a BKK Query."""
        transID = self.paramValues["TransformationID"]
        if self.exists and transID:
            res = self.transClient.deleteTransformationBookkeepingQuery(transID)
            if not res["OK"]:
                return res
        self.item_called = "BkQuery"
        self.paramValues[self.item_called] = {}
        return S_OK()

    #############################################################################

    def addTransformation(self, addFiles=True, printOutput=False):
        """Add a transformation, using TransformationClient()"""
        res = super().addTransformation(addFiles, printOutput)
        if res["OK"]:
            transID = res["Value"]
        else:
            return res

        bkQuery = self.paramValues.get("BkQuery")
        if bkQuery:
            res = self.setBkQuery(bkQuery)
            if not res["OK"]:
                return self._errorReport(res, "Failed to set BK query")
        else:
            self.transClient.deleteTransformationParameter(transID, "BkQuery")

        return S_OK(transID)

    def setSEParam(self, key, seList):
        return self.__setSE(key, seList)

    def setAdditionalParam(self, key, val):
        self.item_called = key
        return self.__setParam(val)

    # This is a trick to overwrite the __checkSEs method of the base class
    def _Transformation__checkSEs(self, seList):
        # This test allows to set some parameters empty
        if seList == []:
            return S_OK()
        if resolveSEGroup(seList):
            return S_OK()
        gLogger.error(f"Some SEs are unknown in {','.join(seList)}")
        return S_ERROR("Some StorageElements not known")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCb API Class.

The LHCb API exposes LHCb specific functionality in addition to the
standard DIRAC API.
"""
import os
import time

from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites
from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite, getSitesForSE
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus

from LHCbDIRAC.Core.Utilities.File import makeGuid
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery
from LHCbDIRAC.DataManagementSystem.Client.DMScript import printDMResult
from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import getAccessURL

COMPONENT_NAME = "DiracLHCb"


def getSiteForSE(se):
    """Get site name for the given SE."""
    result = getSitesForSE(se)
    if not result["OK"]:
        return result
    if result["Value"]:
        return S_OK(result["Value"][0])
    return S_OK("")


def translateBKPath(bkPath, procPassID=3):
    bk = [s for s in bkPath.split("/") if s]
    if procPassID < 0:
        return bk
    try:
        bkNodes = bk[0:procPassID]
        bkNodes.append("/" + "/".join(bk[procPassID:-2]))
        bkNodes.append(bk[-2])
        bkNodes.append(bk[-1])
    except Exception:
        gLogger.error("Incorrect BKQuery")
        bkNodes = None
    return bkNodes


class DiracLHCb(Dirac):
    #############################################################################
    def __init__(self, withRepo=False, repoLocation="", operationsHelperIn=None):
        """Internal initialization of the DIRAC API."""

        super().__init__(withRepo=withRepo, repoLocation=repoLocation)
        self.tier1s = []

        if not operationsHelperIn:
            self.opsH = Operations()
        else:
            self.opsH = operationsHelperIn

        self._bkQueryTemplate = {
            "SimulationConditions": "All",
            "DataTakingConditions": "All",
            "ProcessingPass": "All",
            "FileType": "All",
            "EventType": 0,
            "ConfigName": "All",
            "ConfigVersion": "All",
            "Production": 0,
            "StartRun": 0,
            "EndRun": 0,
            "DataQuality": "All",
            "Visible": "Yes",
            "ExtendedDQOK": "All",
            "SMOG2": "All",
        }
        self._bkClient = BookkeepingClient()  # to expose all BK client methods indirectly

    #############################################################################
    def addRootFile(self, lfn, fullPath, diracSE, printOutput=False):
        """Add a Root file to Grid storage, an attempt is made to retrieve the POOL
        GUID of the file prior to upload.

           Example Usage:

           >>> print dirac.addFile('/lhcb/user/p/paterson/myRootFile.tar.gz','myFile.tar.gz','CERN-USER')
           {'OK': True, 'Value':{'Failed': {},
            'Successful': {'/lhcb/user/p/paterson/test/myRootFile.tar.gz': {'put': 64.246301889419556,
                                                                        'register': 1.1102778911590576}}}}

           @param lfn: Logical File Name (LFN)
           @type lfn: string
           @param diracSE: DIRAC SE name e.g. CERN-USER
           @type diracSE: strin
           @param printOutput: Optional flag to print result
           @type printOutput: boolean
           @return: S_OK,S_ERROR
        """
        return super().addFile(lfn, fullPath, diracSE, fileGuid=makeGuid(fullPath)[fullPath], printOutput=printOutput)

    def addFile(self, lfn, fullPath, diracSE, printOutput=False):  # pylint: disable=arguments-differ
        """Copy of addRootFile."""
        return super().addFile(lfn, fullPath, diracSE, fileGuid=makeGuid(fullPath)[fullPath], printOutput=printOutput)

    def getBKAncestors(self, lfns, depth=1, replica=True):
        """This function allows to retrieve ancestor files from the Bookkeeping.

         Example Usage:

         >>> dirac.getBKAncestors('/lhcb/data/2009/DST/00005727/0000/00005727_00000042_1.dst',2)
         {'OK': True, 'Value': ['/lhcb/data/2009/DST/00005727/0000/00005727_00000042_1.dst',
         '/lhcb/data/2009/RAW/FULL/LHCb/COLLISION09/63807/063807_0000000004.raw']}

        @param lfn: Logical File Name (LFN)
        @type lfn: string or list
        @param depth: Ancestor depth
        @type depth: integer
        """

        result = self._bkClient.getFileAncestors(lfns, depth, replica=replica)
        if not result["OK"]:
            self.log.error("Could not get ancestors", result["Message"])
            return result
        ancestors = {x["FileName"] for ancestors in result["Value"]["Successful"].values() for x in ancestors}

        return S_OK(lfns + list(ancestors))

    #############################################################################
    def bkQueryRunsByDate(
        self, bkPath, startDate, endDate, dqFlag="All", selection="Runs", extendedDQOK="All", SMOG2="All"
    ):
        """This function allows to create and perform a BK query given a supplied BK path

        The following BK path convention is expected:

        .. code-block:: none

          /<ConfigurationName>/<Configuration Version>/<Condition Description><Processing Pass>/<Event Type>/<File Type>

        so an example for 2016 collisions data would be:

        .. code-block:: none

          /LHCb/Collision09//LHCb/Collision16/Beam6500GeV-VeloClosed-MagDown/Real Data/Reco16/Stripping26/90000000/EW.DST

        The startDate and endDate must be specified as yyyy-mm-dd.

        Runs can be selected based on their status e.g. the selection parameter
        has the following possible attributes:

         - Runs - data for all runs in the range are queried (default)
         - ProcessedRuns - data is retrieved for runs that are processed
         - NotProcessed - data is retrieved for runs that are not yet processed.

        Example Usage:

        .. code-block:: python

          >>> dirac.bkQueryRunsByDate('/LHCb/Collision16//Real Data/90000000/RAW',
                                      '2016-08-20','2016-08-22',dqFlag='OK',selection='Runs')
          {'OK': True, 'Value': [<LFN1>,<LFN2>]}

          >>> dirac.bkQueryRunsByDate('/LHCb/Collision16/Beam6500GeV-VeloClosed-MagDown/Real'
                                      'Data/Reco16/Stripping26/90000000/EW.DST',
                                      '2016-08-20','2016-08-22',dqFlag='OK',selection='Runs')

        @param bkPath: BK path as described above
        @type bkPath: string
        @param dqFlag: Optional Data Quality flag
        @type dqFlag: string
        @param startDate: Start date  yyyy-mm-dd
        @param startDate: string
        @param endDate: End date  yyyy-mm-dd
        @param endDate: string
        @param selection: Either Runs, ProcessedRuns or NotProcessed
        @param selection: string
        @param extendedDQOK: Optional Extended system name(s) which must be ok
        @type extendedDQOK: string | string[]
        @param SMOG2: Optional SMOG2 states (any from the list)
        @type SMOG2: string | string[]
        @return: S_OK,S_ERROR
        """
        runSelection = ["Runs"]
        if selection not in runSelection:
            return S_ERROR(f"Expected one of {', '.join(runSelection)} not \"{selection}\" for selection")

        if not isinstance(bkPath, str):
            return S_ERROR("Expected string for bkPath")

        # remove any double slashes, spaces must be preserved
        # remove any empty components from leading and trailing slashes
        bkQuery = BKQuery().buildBKQuery(bkPath)
        if not bkQuery:
            return S_ERROR(
                "Please provide a BK path: "
                "/<ConfigurationName>/<Configuration Version>/<Condition Description>/<Processing Pass>"
                "/<Event Type>/<File Type>"
            )

        if not startDate or not endDate:
            return S_ERROR("Expected both start and end dates to be defined in format: yyyy-mm-dd")

        if not isinstance(startDate, str) or not isinstance(endDate, str):
            return S_ERROR("Expected yyyy-mm-dd string for start and end dates")

        if not len(startDate.split("-")) == 3 or not len(endDate.split("-")) == 3:
            return S_ERROR("Expected yyyy-mm-dd string for start and end dates")

        start = time.time()
        result = self._bkClient.getRunsForAGivenPeriod({"StartDate": startDate, "EndDate": endDate})
        rtime = time.time() - start
        self.log.info(f"BK query time: {rtime:.2f} sec")
        if not result["OK"]:
            self.log.info(f'Could not get runs with given dates from BK with result: "{result}"')
            return result

        if not result["Value"]:
            self.log.info("No runs selected from BK for specified dates")
            return result

        if selection not in result["Value"]:
            return S_ERROR(f"No {selection} runs for specified dates")

        runs = result["Value"][selection]
        self.log.info(f"Found the following {len(runs)} runs:\n{', '.join([str(i) for i in runs])}")
        # temporary until we can query for a discrete list of runs
        selectedData = []
        for run in runs:
            query = bkQuery.copy()
            query["StartRun"] = run
            query["EndRun"] = run
            if dqFlag:
                check = self.__checkDQFlags(dqFlag)
                if not check["OK"]:
                    return check
                dqFlag = check["Value"]
                query["DataQuality"] = dqFlag
            if extendedDQOK:
                check = self.__checkExtendedDQOK(dqFlag, extendedDQOK)
                if not check["OK"]:
                    return check
                extendedDQOK = check["Value"]
                query["ExtendedDQOK"] = extendedDQOK
            if SMOG2:
                check = self.__checkSMOG2(SMOG2)
                if not check["OK"]:
                    return check
                SMOG2 = check["Value"]
                query["SMOG2"] = SMOG2
            start = time.time()
            result = self._bkClient.getVisibleFilesWithMetadata(query)
            rtime = time.time() - start
            self.log.info(f"BK query time: {rtime:.2f} sec")
            self.log.verbose(result)
            if not result["OK"]:
                return result
            self.log.info(f"Selected {len(result['Value'])} files for run {run}")
            if result["Value"]["LFNs"]:
                selectedData += list(result["Value"]["LFNs"])

        self.log.info(f"Total files selected = {len(selectedData)}")
        return S_OK(selectedData)

    #############################################################################
    def bkQueryRun(self, bkPath, dqFlag="All", extendedDQOK="All", SMOG2="All"):
        """This function allows to create and perform a BK query given a supplied
        BK path. The following BK path convention is expected:

            /<Run Number>/<Processing Pass>/<Event Type>/<File Type>

            so an example for 2009 collisions data would be:

           /63566/Real Data + RecoToDST-07/90000000/DST

           In addition users can specify a range of runs using the following convention:

           /<Run Number 1> - <Run Number 2>/<Processing Pass>/<Event Type>/<File Type>

           so extending the above example this would look like:

           /63566-63600/Real Data + RecoToDST-07/90000000/DST

           Example Usage:

           >>> dirac.bkQueryRun('/63566/Real Data/RecoToDST-07/90000000/DST')
           {'OK':True,'Value': ['/lhcb/data/2009/DST/00005842/0000/00005842_00000008_1.dst']}

           @param bkPath: BK path as described above
           @type bkPath: string
           @param dqFlag: Optional Data Quality flag
           @type dqFlag: string
           @param extendedDQOK: Optional Extended system name(s) which must be ok
           @type extendedDQOK: string | string[]
           @param SMOG2: Optional SMOG2 states (any from the list)
           @type SMOG2: string | string[]
           @return: S_OK,S_ERROR
        """
        if not isinstance(bkPath, str):
            return S_ERROR("Expected string for bkPath")

        # remove any double slashes, spaces must be preserved
        # remove any empty components from leading and trailing slashes
        bkPath = translateBKPath(bkPath, procPassID=1)
        if not len(bkPath) == 4:
            return S_ERROR(
                "Expected 4 components to the BK path: /<Run Number>/<Processing Pass>/<Event Type>/<File Type>"
            )

        runNumberString = bkPath[0].replace("--", "-").replace(" ", "")
        startRun = 0
        endRun = 0
        if "-" in runNumberString:
            runs = runNumberString.split("-")
            if len(runs) != 2:
                return S_ERROR(f'Could not determine run range from "{runNumberString}", try "<Run 1> - <Run2>"')
            try:
                start = int(runs[0])
                end = int(runs[1])
            except Exception:
                return S_ERROR(f"Invalid run range: {runNumberString}")
            startRun = min(start, end)
            endRun = max(start, end)
        else:
            try:
                startRun = int(runNumberString)
                endRun = startRun
            except Exception:
                return S_ERROR(f"Invalid run number: {runNumberString}")

        query = self._bkQueryTemplate.copy()
        query["StartRun"] = startRun
        query["EndRun"] = endRun
        query["ProcessingPass"] = bkPath[1]
        query["EventType"] = bkPath[2]
        query["FileType"] = bkPath[3]

        if dqFlag:
            check = self.__checkDQFlags(dqFlag)
            if not check["OK"]:
                return check
            dqFlag = check["Value"]
            query["DataQuality"] = dqFlag
        if extendedDQOK:
            check = self.__checkExtendedDQOK(dqFlag, extendedDQOK)
            if not check["OK"]:
                return check
            extendedDQOK = check["Value"]
            query["ExtendedDQOK"] = extendedDQOK
        if SMOG2:
            check = self.__checkSMOG2(SMOG2)
            if not check["OK"]:
                return check
            SMOG2 = check["Value"]
            query["SMOG2"] = SMOG2
        result = self.bkQuery(query)
        self.log.verbose(result)
        return result

    #############################################################################
    def bkQueryProduction(self, bkPath, dqFlag="All", extendedDQOK="All", SMOG2="All"):
        """This function allows to create and perform a BK query given a supplied
        BK path. The following BK path convention is expected:

            /<ProductionID>/[<Processing Pass>/<Event Type>/]<File Type>

            so an example for 2009 collisions data would be:

           /5842/Real Data + RecoToDST-07/90000000/DST

           Note that neither the processing pass nor the event type should be necessary. So either of them can be omitted

           a data quality flag can also optionally be provided, the full list of these is available
           via the getAllDQFlags() method.

           Example Usage:

           >>> dirac.bkQueryProduction('/5842/Real Data/RecoToDST-07/90000000/DST')
           {'OK': True, 'Value': [<LFN1>,<LFN2>]}

           @param bkPath: BK path as described above
           @type bkPath: string
           @param dqFlag: Optional Data Quality flag
           @type dqFlag: string
           @param extendedDQOK: Optional Extended system name(s) which must be ok
           @type extendedDQOK: string | string[]
           @param SMOG2: Optional SMOG2 states (any from the list)
           @type SMOG2: string | string[]
           @return: S_OK,S_ERROR
        """
        if not isinstance(bkPath, str):
            return S_ERROR("Expected string for bkPath")

        # remove any double slashes, spaces must be preserved
        # remove any empty components from leading and trailing slashes
        bkPath = translateBKPath(bkPath, procPassID=1)
        if len(bkPath) < 2:
            return S_ERROR("Invalid bkPath: should at least contain /ProductionID/FileType")
        query = self._bkQueryTemplate.copy()
        try:
            query["Production"] = int(bkPath[0])
        except Exception:
            return S_ERROR("Invalid production ID")
        query["FileType"] = bkPath[-1]

        if dqFlag:
            check = self.__checkDQFlags(dqFlag)
            if not check["OK"]:
                return check
            dqFlag = check["Value"]
            query["DataQuality"] = dqFlag
        if extendedDQOK:
            check = self.__checkExtendedDQOK(dqFlag, extendedDQOK)
            if not check["OK"]:
                return check
            extendedDQOK = check["Value"]
            query["ExtendedDQOK"] = extendedDQOK
        if SMOG2:
            check = self.__checkSMOG2(SMOG2)
            if not check["OK"]:
                return check
            SMOG2 = check["Value"]
            query["SMOG2"] = SMOG2

        for key, val in list(query.items()):
            if isinstance(val, str) and val.lower() == "all":
                query.pop(key)
        result = self.bkQuery(query)
        self.log.verbose(result)
        return result

    #############################################################################
    def bkQueryPath(self, bkPath, dqFlag="All", extendedDQOK="All", SMOG2="All"):
        """This function allows to create and perform a BK query given a supplied
        BK path. The following BK path convention is expected:

           /<ConfigurationName>/<Configuration Version>/<Sim or Data Taking Condition>
           /<Processing Pass>/<Event Type>/<File Type>

           so an example for 2009 collsions data would be:

           /LHCb/Collision09/Beam450GeV-VeloOpen-MagDown/Real Data + RecoToDST-07/90000000/DST

           or for MC09 simulated data:

           /MC/2010/Beam3500GeV-VeloClosed-MagDown-Nu1/2010-Sim01Reco01-withTruth/27163001/DST

           a data quality flag can also optionally be provided, the full list of these is available
           via the getAllDQFlags() method.

           Example Usage:

           >>> dirac.bkQueryPath('/MC/2010/Beam3500GeV-VeloClosed-MagDown-Nu1/Sim07/Reco06-withTruth/10012004/DST')
           {'OK': True, 'Value': [<LFN1>,<LFN2>]}

           @param bkPath: BK path as described above
           @type bkPath: string
           @param dqFlag: Optional Data Quality flag
           @type dqFlag: string
           @param extendedDQOK: Optional Extended system name(s) which must be ok
           @type extendedDQOK: string | string[]
           @param SMOG2: Optional SMOG2 states (any from the list)
           @type SMOG2: string | string[]
           @return: S_OK,S_ERROR
        """
        if not isinstance(bkPath, str):
            return S_ERROR("Expected string for bkPath")

        # remove any double slashes, spaces must be preserved
        # remove any empty components from leading and trailing slashes
        bkPath = translateBKPath(bkPath, procPassID=3)
        if not len(bkPath) == 6:
            return S_ERROR(
                "Expected 6 components to the BK path: "
                "/<ConfigurationName>/<Configuration Version>/<Sim or Data Taking Condition>"
                "/<Processing Pass>/<Event Type>/<File Type>"
            )

        query = self._bkQueryTemplate.copy()
        query["ConfigName"] = bkPath[0]
        query["ConfigVersion"] = bkPath[1]
        query["ProcessingPass"] = bkPath[3]
        query["EventType"] = bkPath[4]
        query["FileType"] = bkPath[5]

        if dqFlag:
            check = self.__checkDQFlags(dqFlag)
            if not check["OK"]:
                return check
            dqFlag = check["Value"]
            query["DataQuality"] = dqFlag
        if extendedDQOK:
            check = self.__checkExtendedDQOK(dqFlag, extendedDQOK)
            if not check["OK"]:
                return check
            extendedDQOK = check["Value"]
            query["ExtendedDQOK"] = extendedDQOK
        if SMOG2:
            check = self.__checkSMOG2(SMOG2)
            if not check["OK"]:
                return check
            SMOG2 = check["Value"]
            query["SMOG2"] = SMOG2

        # The problem here is that we don't know if it's a sim or data taking condition,
        # assume that if configName=MC this is simulation
        if bkPath[0].lower() == "mc":
            query["SimulationConditions"] = bkPath[2]
        else:
            query["DataTakingConditions"] = bkPath[2]

        result = self.bkQuery(query)
        self.log.verbose(result)
        return result

    #############################################################################
    def bookkeepingQuery(
        self,
        SimulationConditions="All",
        DataTakingConditions="All",
        ProcessingPass="All",
        FileType="All",
        EventType=0,
        ConfigName="All",
        ConfigVersion="All",
        ProductionID=0,
        DataQuality="ALL",
        ExtendedDQOK="ALL",
        SMOG2="ALL",
    ):
        """This function will create and perform a BK query using the supplied
        arguments and return a list of LFNs.

            Example Usage:

            >>> dirac.bookkeepingQuery(ConfigName='LHCb',ConfigVersion='Collision09',
            EventType='90000000',ProcessingPass='Real Data',DataTakingConditions='Beam450GeV-VeloOpen-MagDown')
            {'OK':True,'Value':<files>}

           @param  ConfigName: BK ConfigName
           @type ConfigName: string
           @param  EventType: BK EventType
           @type EventType: string
           @param  FileType: BK FileType
           @type FileType: string
           @param  ProcessingPass: BK ProcessingPass
           @type ProcessingPass: string
           @param  ProductionID: BK ProductionID
           @type ProductionID: integer
           @param  DataQuality: BK DataQuality
           @type DataQuality: string
           @param  ConfigVersion: BK ConfigVersion
           @type ConfigVersion: string
           @param  DataTakingConditions: BK DataTakingConditions
           @type DataTakingConditions: string
           @param  SimulationConditions: BK SimulationConditions
           @type SimulationConditions: string
           @param ExtendedDQOK: BK extended DQ OK flag(s), comma separated
           @type ExtendedDQOK: string
           @param SMOG2: BK SMOG2 state(s), comma separated
           @type SMOG2: string
           @return: S_OK,S_ERROR
        """
        query = self._bkQueryTemplate.copy()
        query["SimulationConditions"] = SimulationConditions
        query["DataTakingConditions"] = DataTakingConditions
        query["ProcessingPass"] = ProcessingPass
        query["FileType"] = FileType
        query["EventType"] = EventType
        query["ConfigName"] = ConfigName
        query["ConfigVersion"] = ConfigVersion
        query["Production"] = ProductionID
        query["DataQuality"] = DataQuality
        query["ExtendedDQOK"] = ExtendedDQOK
        query["SMOG2"] = SMOG2
        return self.bkQuery(query)

    #############################################################################
    def bkQuery(self, bkQueryDict):
        """Developer function. Perform a query to the LHCb Bookkeeping to return a
        list of LFN(s). This method takes a BK query dictionary.

            Example Usage:

            >>> print dirac.bkQuery(query)
            {'OK':True,'Value':<files>}

           @param bkQueryDict: BK query
           @type bkQueryDict: dictionary (see bookkeepingQuery() for keys)
           @return: S_OK,S_ERROR
        """
        problematicFields = []
        # Remove the Visible flag as anyway the method is for visible files ;-)
        # bkQueryDict.setdefault( 'Visible', 'Yes' )
        for name, value in bkQueryDict.items():
            if name not in self._bkQueryTemplate:
                problematicFields.append(name)

        if problematicFields:
            msg = "The following fields are not valid for a BK query: {}\nValid fields include: {}".format(
                ", ".join(problematicFields),
                ", ".join(self._bkQueryTemplate),
            )
            return S_ERROR(msg)

        for name, value in list(bkQueryDict.items()):
            if name == "Production" or name == "EventType" or name == "StartRun" or name == "EndRun":
                try:
                    ivalue = int(value)
                except ValueError:
                    return S_ERROR(f"{name} must be a number, got '{value}'")
                if ivalue == 0:
                    del bkQueryDict[name]
                else:
                    bkQueryDict[name] = str(ivalue)
            elif name == "FileType":
                if value.lower() == "all":
                    bkQueryDict[name] = "ALL"
            else:
                if str(value).lower() == "all":
                    del bkQueryDict[name]

        if "Production" in bkQueryDict or "StartRun" in bkQueryDict or "EndRun" in bkQueryDict:
            self.log.verbose("Found a specific query so loosening some restrictions to prevent BK overloading")
        else:
            if "SimulationConditions" not in bkQueryDict and "DataTakingConditions" not in bkQueryDict:
                return S_ERROR("A Simulation or DataTaking Condition must be specified for a BK query.")
            if (
                "EventType" not in bkQueryDict
                and "ConfigName" not in bkQueryDict
                and "ConfigVersion" not in bkQueryDict
            ):
                return S_ERROR(
                    "The minimal set of BK fields for a query is: EventType, ConfigName and ConfigVersion"
                    " in addition to a Simulation or DataTaking Condition"
                )

        self.log.verbose("Final BK query dictionary is:")
        for item in bkQueryDict.items():
            self.log.verbose("%s : %s" % item)

        start = time.time()
        result = self._bkClient.getVisibleFilesWithMetadata(bkQueryDict)
        #    result = bk.getFilesWithGivenDataSets(bkQueryDict)
        rtime = time.time() - start
        self.log.info(f"BK query time: {rtime:.2f} sec")

        if not result["OK"]:
            return S_ERROR(f"BK query returned an error: \"{result['Message']}\"")

        if not result["Value"]["LFNs"]:
            return self._errorReport("No BK files selected")

        returnedFiles = result["Value"]["Summary"]["Number Of Files"]
        self.log.verbose(f"{returnedFiles} files selected from the BK")
        return result

    #############################################################################
    def __checkDQFlags(self, flags):
        """Internal function.

        Checks the provided flags against the list of possible DQ flag
        statuses from the Bookkeeping.
        """
        dqFlags = []
        if isinstance(flags, list):
            dqFlags = flags
        else:
            dqFlags = [flags]

        bkFlags = self.getAllDQFlags()
        if not bkFlags["OK"]:
            return bkFlags

        final = []
        for flag in dqFlags:
            if flag.lower() == "all":
                final.append(flag.upper())
            else:
                flag = flag.upper()
                if flag not in bkFlags["Value"]:
                    msg = f"Specified DQ flag \"{flag}\" is not in allowed list: {', '.join(bkFlags['Value'])}"
                    self.log.error(msg)
                    return S_ERROR(msg)
                else:
                    final.append(flag)

        # when first coding this it was not possible to use a list ;)
        if len(final) == 1:
            final = final[0]

        return S_OK(final)

    #############################################################################
    def getAllDQFlags(self, printOutput=False):
        """Helper function.  Returns the list of possible DQ flag statuses from the
        Bookkeeping.

            Example Usage:

            >>> print dirac.getAllDQFlags()
            {'OK':True,'Value':<flags>}

           @param printOutput: Optional flag to print result
           @type printOutput: boolean
           @return: S_OK,S_ERROR
        """
        result = self._bkClient.getAvailableDataQuality()
        if not result["OK"]:
            self.log.error(f"Could not obtain possible DQ flags from BK with result:\n{result}")
            return result

        if printOutput:
            flags = result["Value"]
            self.log.info(f"Possible DQ flags from BK are: {', '.join(flags)}")

        return result

    #############################################################################
    def __checkExtendedDQOK(self, flags, systems):
        """Internal function.

        Checks the provided systems against the list of possible Extended DQOK systems
        in the Bookkeeping.
        """
        extDQOK = systems if isinstance(systems, list) else [systems]
        if len(extDQOK) == 1 and extDQOK[0].lower() == "all":
            return S_OK("ALL")  # technically ANY
        dqFlags = flags if isinstance(flags, list) else [flags]
        if len(dqFlags) != 1 or dqFlags[0].lower() != "ok":
            msg = "DQ flag 'OK' must be expicitly specified when using Extended DQOK"
            self.log.error(msg)
            return S_ERROR(msg)
        bkExtDQOK = self.getAllExtendedDQOK()
        if not bkExtDQOK["OK"]:
            return bkExtDQOK

        for system in extDQOK:
            if system not in bkExtDQOK["Value"]:
                msg = f"Specified Extended DQOK system \"{system}\" is not in allowed list: {', '.join(bkExtDQOK['Value'])}"
                self.log.error(msg)
                return S_ERROR(msg)

        return S_OK(extDQOK)

    #############################################################################
    def getAllExtendedDQOK(self, printOutput=False):
        """Helper function.  Returns the list of possible ExtendedDQOK systems from the
        Bookkeeping.

            Example Usage:

            >>> print dirac.getAllExtendedDQOK()
            {'OK':True,'Value':<systems>}

           @param printOutput: Optional flag to print result
           @type printOutput: boolean
           @return: S_OK,S_ERROR
        """
        result = self._bkClient.getAvailableExtendedDQOK()
        if not result["OK"]:
            self.log.error(f"Could not obtain possible Extended DQOK systems from BK with result:\n{result}")
            return result

        if printOutput:
            systemnames = result["Value"]
            self.log.info(f"Possible ExtendedDQOK systems from BK are: {', '.join(systemnames)}")

        return result

    #############################################################################
    def __checkSMOG2(self, states):
        """Internal function.

        Checks provided SMOG2 states against the list of possible SMOG2 states
        in the Bookkeeping.
        """
        smog2States = states if isinstance(states, list) else [states]
        if len(smog2States) == 1 and smog2States[0].lower() == "all":
            return S_OK("ALL")  # technically ANY
        bkSMOG2States = self.getAllSMOG2States()
        if not bkSMOG2States["OK"]:
            return bkSMOG2States

        for state in smog2States:
            if state not in bkSMOG2States["Value"]:
                msg = f"Specified SMOG2 state \"{state}\" is not in allowed list: {', '.join(bkSMOG2States['Value'])}"
                self.log.error(msg)
                return S_ERROR(msg)

        return S_OK(smog2States)

    #############################################################################
    def getAllSMOG2States(self, printOutput=False):
        """Helper function.  Returns the list of possible SMOG2 states from the Bookkeeping.

         Example Usage:

         >>> print dirac.getAllSMOG2States()
         {'OK':True,'Value':<states>}

        @param printOutput: Optional flag to print result
        @type printOutput: boolean
        @return: S_OK,S_ERROR
        """
        result = self._bkClient.getAvailableSMOG2States()
        if not result["OK"]:
            self.log.error(f"Could not obtain possible SMOG2 states from BK with result:\n{result}")
            return result

        if printOutput:
            states = result["Value"]
            self.log.info(f"Possible SMOG2 states from BK are: {', '.join(states)}")

        return result

    #############################################################################
    def getDataByRun(self, lfns, printOutput=False):
        """Sort the supplied lfn list by run. An S_OK object will be returned
        containing a dictionary of runs and the corresponding list of LFN(s)
        associated with them.

           Example usage:

           >>> print dirac.getDataByRun(lfns)
           {'OK': True, 'Value': {<RUN>:['<LFN>','<LFN>',...], <RUN>:['<LFN>',..]}}


           @param lfns: Logical File Name(s)
           @type lfns: list
           @param printOutput: Optional flag to print result
           @type printOutput: boolean
           @return: S_OK,S_ERROR
        """
        if isinstance(lfns, str):
            lfns = [lfns.replace("LFN:", "")]
        elif isinstance(lfns, list):
            try:
                lfns = [str(lfn.replace("LFN:", "")) for lfn in lfns]
            except ValueError as x:
                return self._errorReport(str(x), "Expected strings for LFNs")
        else:
            return self._errorReport("Expected single string or list of strings for LFN(s)")

        runDict = {}
        start = time.time()
        result = self._bkClient.getFileMetadata(lfns)
        self.log.verbose(f"Obtained BK file metadata in {time.time() - start:.2f} seconds")
        if not result["OK"]:
            self.log.error(f"Failed to get bookkeeping metadata with result \"{result['Message']}\"")
            return result

        for lfn, metadata in result["Value"]["Successful"].items():
            if "RunNumber" in metadata:
                runNumber = metadata["RunNumber"]
                runDict.setdefault(runNumber, []).append(lfn)
            else:
                self.log.warn(f"Could not find run number from BK for {lfn}")

        if printOutput:
            self.log.notice(self.pPrint.pformat(runDict))

        return S_OK(runDict)

    #############################################################################
    def bkMetadata(self, lfns, printOutput=False):
        """Return metadata for the supplied lfn list. An S_OK object will be
        returned containing a dictionary of LFN(s) and the corresponding metadata
        associated with them.

           Example usage:

           >>> print dirac.bkMetadata(lfns)
           {'OK': True, 'Value': {<LFN>:{'<Name>':'<Value>',...},...}}

           @param lfns: Logical File Name(s)
           @type lfns: list
           @param printOutput: Optional flag to print result
           @type printOutput: boolean
           @return: S_OK,S_ERROR
        """
        if isinstance(lfns, str):
            lfns = [lfns.replace("LFN:", "")]
        elif isinstance(lfns, list):
            try:
                lfns = [str(lfn.replace("LFN:", "")) for lfn in lfns]
            except ValueError as x:
                return self._errorReport(str(x), "Expected strings for LFNs")
        else:
            return self._errorReport("Expected single string or list of strings for LFN(s)")

        start = time.time()
        result = self._bkClient.getFileMetadata(lfns)
        self.log.verbose(f"Obtained BK file metadata in {time.time() - start:.2f} seconds")
        if not result["OK"]:
            self.log.error(f"Failed to get bookkeeping metadata with result \"{result['Message']}\"")
            return result

        if printOutput:
            self.log.notice(self.pPrint.pformat(result["Value"]))

        return result

    #############################################################################

    def lhcbProxyInit(self, *args):  # pylint: disable=no-self-use
        """just calling the dirac-proxy-init script."""
        os.system("dirac-proxy-init -o LogLevel=NOTICE -t %s" % "' '".join(args))

    #############################################################################

    def lhcbProxyInfo(self, *args):  # pylint: disable=no-self-use
        """just calling the dirac-proxy-info script."""
        os.system("dirac-proxy-info -o LogLevel=NOTICE %s" % "' '".join(args))

    #############################################################################

    def gridWeather(self, printOutput=False):
        """This method gives a snapshot of the current Grid weather from the
        perspective of the DIRAC site and SE masks.  Tier-1 sites are returned with
        more detailed information.

           Example usage:

           >>> print dirac.gridWeather()
           {'OK': True, 'Value': {{'Sites':<siteInfo>,'SEs':<seInfo>,'Tier-1s':<tierInfo>}}

           @param printOutput: Optional flag to print result
           @type printOutput: boolean
           @return: S_OK,S_ERROR
        """

        lcgSites = gConfig.getSections("/Resources/Sites/LCG")
        if not lcgSites["OK"]:
            return lcgSites

        for lcgSite in lcgSites["Value"]:
            tier = gConfig.getValue(f"/Resources/Sites/LCG/{lcgSite}/MoUTierLevel", 2)
            if tier in (0, 1):
                self.tier1s.append(lcgSite)

        siteInfo = self.checkSites()
        if not siteInfo["OK"]:
            return siteInfo
        siteInfo = siteInfo["Value"]

        seInfo = self.checkSEs()
        if not seInfo["OK"]:
            return seInfo
        seInfo = seInfo["Value"]

        tierSEs = {}
        for site in self.tier1s:
            tierSEs[site] = getSEsForSite(site)["Value"]

        tierInfo = {}
        for site, seList in tierSEs.items():
            tierInfo[site] = {}
            for se in seList:
                if se in seInfo:
                    tierSEInfo = seInfo[se]
                    tierInfo[site][se] = tierSEInfo
            if site in siteInfo["AllowedSites"]:
                tierInfo[site]["MaskStatus"] = "Allowed"
            else:
                tierInfo[site]["MaskStatus"] = "Banned"

        if printOutput:
            self.log.notice("========> Tier-1 status in DIRAC site and SE masks")
            for site in sorted(self.tier1s):
                self.log.notice(f"\n====> {site} is {tierInfo[site]['MaskStatus']} in site mask\n")
                self.log.notice(f"{'Storage Element'.ljust(25)} {'Read Status'.rjust(15)} {'Write Status'.rjust(15)}")
                for se in sorted(tierSEs[site]):
                    if se in tierInfo[site]:
                        self.log.notice(
                            "%s %s %s"
                            % (
                                se.ljust(25),
                                tierInfo[site][se]["ReadStatus"].rjust(15),
                                tierInfo[site][se]["WriteStatus"].rjust(15),
                            )
                        )

            self.log.notice("\n========> Tier-2 status in DIRAC site mask\n")
            allowedSites = siteInfo["AllowedSites"]
            bannedSites = siteInfo["BannedSites"]
            for site in self.tier1s:
                if site in allowedSites:
                    allowedSites.remove(site)
                if site in bannedSites:
                    bannedSites.remove(site)
            self.log.notice(f" {len(allowedSites)} sites are in the site mask, {len(bannedSites)} are banned.\n")

        summary = {"Sites": siteInfo, "SEs": seInfo, "Tier-1s": tierInfo}
        return S_OK(summary)

    #############################################################################
    def checkSites(self, printOutput=False):  # pylint: disable=no-self-use
        """Return the list of sites in the DIRAC site mask and those which are
        banned.

           Example usage:

           >>> print dirac.checkSites()
           {'OK': True, 'Value': {'AllowedSites':['<Site>',...],'BannedSites':[]}

           @param printOutput: Optional flag to print result
           @type printOutput: boolean
           @return: S_OK,S_ERROR
        """

        res = getSites()
        if not res["OK"]:
            self.log.error("Could not get list of sites from CS", res["Message"])
            return res
        totalList = res["Value"]

        res = DiracAdmin().getSiteMask()
        if not res["OK"]:
            return res

        sites = res["Value"]
        bannedSites = []
        for site in totalList:
            if site not in sites:
                bannedSites.append(site)

        if printOutput:
            self.log.notice("\n========> Allowed Sites\n")
            self.log.notice("\n".join(sites))
            self.log.notice("\n========> Banned Sites\n")
            self.log.notice("\n".join(bannedSites))
            self.log.notice(
                "\nThere is a total of %s allowed sites and %s banned sites in the system."
                % (len(sites), len(bannedSites))
            )

        return S_OK({"AllowedSites": sites, "BannedSites": bannedSites})

    #############################################################################
    def checkSEs(self, printOutput=False):  # pylint: disable=no-self-use
        """Check the status of read and write operations in the DIRAC SE mask.

        Example usage:

        >>> print dirac.checkSEs()
        {'OK': True, 'Value': {<LFN>:{'<Name>':'<Value>',...},...}}

        @param printOutput: Optional flag to print result
        @type printOutput: boolean
        @return: S_OK,S_ERROR
        """
        res = gConfig.getSections("/Resources/StorageElements", True)

        if not res["OK"]:
            self.log.error("Failed to get storage element information", res["Message"])
            return res

        if printOutput:
            self.log.notice(f"{'Storage Element'.ljust(25)} {'Read Status'.rjust(15)} {'Write Status'.rjust(15)}")

        seList = sorted(res["Value"])
        result = {}
        rss = ResourceStatus()
        for se in seList:
            res = rss.getElementStatus(se, "StorageElement")
            if not res["OK"]:
                self.log.error(f"Failed to get StorageElement status for {se}")
            else:
                readState = res["Value"].get("ReadAccess", "Active")
                writeState = res["Value"].get("WriteAccess", "Active")
                result[se] = {"ReadStatus": readState, "WriteStatus": writeState}
                if printOutput:
                    self.log.notice(f"{se.ljust(25)} {readState.rjust(15)} {writeState.rjust(15)}")

        return S_OK(result)

    def splitInputDataBySize(self, lfns, maxSizePerJob=20, printOutput=False):
        """Split the supplied lfn list by the replicas present at the possible
        destination sites, based on a maximum size. An S_OK object will be returned
        containing a list of lists in order to create the jobs.

           Example usage:

           >>> d.splitInputDataBySize(lfns,10)
           {'OK': True, 'Value': [['<LFN>'], ['<LFN>']]}


           @param lfns: Logical File Name(s) to split
           @type lfns: list
           @param maxSizePerJob: Maximum size (in GB) per bunch
           @type maxSizePerJob: integer
           @param printOutput: Optional flag to print result
           @type printOutput: boolean
           @return: S_OK,S_ERROR
        """
        sitesForSE = {}
        if isinstance(lfns, str):
            lfns = [lfns.replace("LFN:", "")]
        elif isinstance(lfns, list):
            try:
                lfns = [str(lfn.replace("LFN:", "")) for lfn in lfns]
            except TypeError as x:
                return self._errorReport(str(x), "Expected strings for LFNs")
        else:
            return self._errorReport("Expected single string or list of strings for LFN(s)")

        if not isinstance(maxSizePerJob, int):
            try:
                maxSizePerJob = int(maxSizePerJob)
            except ValueError as x:
                return self._errorReport(str(x), "Expected integer for maxSizePerJob")
        maxSizePerJob *= 1000 * 1000 * 1000

        replicaDict = self.getReplicas(lfns)
        if not replicaDict["OK"]:
            return replicaDict
        replicas = replicaDict["Value"]["Successful"]
        if not replicas:
            return self._errorReport(
                list(replicaDict["Value"]["Failed"].items())[0], "Failed to get replica information"
            )
        siteLfns = {}
        for lfn, reps in replicas.items():
            possibleSites = {
                site for se in reps for site in sitesForSE.setdefault(se, getSitesForSE(se).get("Value", []))
            }
            siteLfns.setdefault(",".join(sorted(possibleSites)), []).append(lfn)

        if "" in siteLfns:
            # Some files don't have active replicas
            return self._errorReport("No active replica found for", str(siteLfns[""]))
        # Get size of files
        metadataDict = self.getLfnMetadata(lfns, printOutput)
        if not metadataDict["OK"]:
            return metadataDict
        fileSizes = {lfn: metadataDict["Value"]["Successful"].get(lfn, {}).get("Size", maxSizePerJob) for lfn in lfns}

        lfnGroups = []
        # maxSize is in GB
        for files in siteLfns.values():
            # Now get bunches of files,
            # Sort in decreasing size
            files.sort(key=fileSizes.__getitem__)
            while files:
                # print [( lfn, fileSizes[lfn] ) for lfn in files]
                group = []
                sizeTot = 0
                for lfn in list(files):
                    size = fileSizes[lfn]
                    if size >= maxSizePerJob:
                        lfnGroups.append([lfn])
                    elif sizeTot + size < maxSizePerJob:
                        sizeTot += size
                        group.append(lfn)
                        files.remove(lfn)
                if group:
                    lfnGroups.append(group)

        if printOutput:
            self.log.notice(self.pPrint.pformat(lfnGroups))
        return S_OK(lfnGroups)

        #############################################################################

    def getAccessURL(self, lfn, storageElement, protocol=None, printOutput=False):
        """Allows to retrieve an access URL for an LFN replica given a valid DIRAC
        SE name.  Contacts the file catalog and contacts the site SRM endpoint
        behind the scenes.

           Example Usage:

           >>> print dirac.getAccessURL('/lhcb/data/CCRC08/DST/00000151/0000/00000151_00004848_2.dst','CERN-RAW')
           {'OK': True, 'Value': {'Successful': {'srm://...': {'SRM2': 'rfio://...'}}, 'Failed': {}}}

           :param lfn: Logical File Name (LFN)
           :type lfn: str or python:list
           :param storageElement: DIRAC SE name e.g. CERN-RAW
           :type storageElement: string
           :param printOutput: Optional flag to print result
           :type printOutput: boolean
           :returns: S_OK,S_ERROR
        """
        ret = self._checkFileArgument(lfn, "LFN")
        if not ret["OK"]:
            return ret
        lfn = ret["Value"]
        if isinstance(lfn, str):
            lfn = [lfn]
        results = getAccessURL(lfn, storageElement, protocol=protocol)
        if printOutput:
            printDMResult(results, empty="File not at SE", script="dirac-dms-lfn-accessURL")
        return results

    #############################################################################

    def _getLocalInputData(self, parameters):
        """LHCb extension of DIRAC API's _getLocalInputData.

        Only used for handling ancestors.
        """
        inputData = parameters.get("InputData")
        if inputData:
            self.log.debug(f"DiracLHCb._getLocalInputData. InputData: {inputData}")
            if isinstance(inputData, str):
                inputData = inputData.split(";")
            inputData = [lfn.strip("LFN:") for lfn in inputData]
            ancestorsDepth = int(parameters.get("AncestorDepth", 0))
            if ancestorsDepth:
                self.log.debug("DiracLHCb._getLocalInputData. ancestorsDepth: %d" % ancestorsDepth)
                res = self._bkClient.getFileAncestors(inputData, ancestorsDepth)
                if not res["OK"]:
                    self.log.error("Can't get ancestors", res["Message"])
                    return res
                ancestorsLFNs = []
                for ancestorsLFN in res["Value"]["Successful"].values():
                    ancestorsLFNs += [i["FileName"] for i in ancestorsLFN]
                self.log.info(f"DiracLHCb._getLocalInputData: adding {len(ancestorsLFNs)} ancestors")
                self.log.verbose("%s", ", ".join(ancestorsLFNs))
                inputData += ancestorsLFNs

        return S_OK(inputData)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""  The MC Simulation Template creates workflows for the following simulation
     use-cases:
      WORKFLOW1: Simulation+Selection+Merge
      WORKFLOW2: Simulation+Selection+MCMerge
      WORKFLOW3: Simulation+Selection
      WORKFLOW4: Simulation+MCMerge
      WORKFLOW5: Simulation

    Exotic things you might want to do:
    * run a local test:
      - of the MC: just set the localTestFlag to True
      - of the merging/stripping: set pr.prodsToLaunch to, e.g., [2], and adjust the pr.inputs at the end of the script
    * run only part of the request on the Grid:
      - for the MC: just set pr.prodsToLaunch = [1]
      - for the merge and/or stripping: set pr.prodsToLaunch, then set pr.previousProdID
"""

import ast

from DIRAC.Core.Base.Script import Script

Script.parseCommandLine()

from DIRAC import gConfig, gLogger, exit as DIRACexit
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import ProductionRequest

#
# Link compression level to the output file visibility. If any file types is visible, the corresponding
# step takes the maximum compression level
# Input : dicts including the visibility of output files per step
# Output: modified list of compression levels
#


def modifyCompression(dict1, dict2, default):
    compression = []

    combinedVis = dict1.copy()
    combinedVis.update(dict2)
    for stepN in sorted(combinedVis):
        if combinedVis[stepN] == "Y":
            compression.append("HIGH")
        else:
            compression.append(default)

    return compression


#
# Simple "check" in case the visibility includes only one element
# Input : dict of visibility
# Output: modified dict of visibility
#


def fillVisList(vdict, num):
    # Assuming that, if there's only one element in the list of output visibility flags, every step will catch that flag
    if len(vdict) == 1:
        val = vdict[list(vdict)[0]]
        vdict = {str(i): val for i in range(int(list(vdict)[0]), int(list(vdict)[0]) + num)}
    # Another assumption: if the number of steps is bigger than that of vis flags,
    # then extend the list with the last flag available
    # to fill the "holes"
    # if len(vlist) < len(slist):
    #  vlist.extend( vlist[-1] * (len(slist) - len(vlist)) )

    return vdict


gLogger = gLogger.getSubLogger("MCSimulation_run.py")
currentSetup = gConfig.getValue("DIRAC/Setup")

pr = ProductionRequest()

stepsList = ["{{p1Step}}"]
stepsList.append("{{p2Step}}")
stepsList.append("{{p3Step}}")
stepsList.append("{{p4Step}}")
stepsList.append("{{p5Step}}")
stepsList.append("{{p6Step}}")
stepsList.append("{{p7Step}}")
stepsList.append("{{p8Step}}")
stepsList.append("{{p9Step}}")
stepsList.append("{{p10Step}}")
stepsList.append("{{p11Step}}")
stepsList.append("{{p12Step}}")
stepsList.append("{{p13Step}}")
stepsList.append("{{p14Step}}")
stepsList.append("{{p15Step}}")
stepsList.append("{{p16Step}}")
stepsList.append("{{p17Step}}")
stepsList.append("{{p18Step}}")
stepsList.append("{{p19Step}}")
stepsList.append("{{p20Step}}")
pr.stepsList = stepsList

###########################################
# Configurable and fixed parameters
###########################################

pr.appendName = "{{WorkflowAppendName#GENERAL: Workflow string to append to production name#1}}"

w = "{{w#----->WORKFLOW: choose one below#}}"
w1 = "{{w1#-WORKFLOW1: Simulation#False}}"
w2 = "{{w2#-WORKFLOW2: Simulation(Gauss) + AllOthers#False}}"
w3 = "{{w3#-WORKFLOW3: Simulation(Gauss) + AllOthers + Merge#False}}"

localTestFlag = "{{localTestFlag#GENERAL: Set True for local test#False}}"
validationFlag = "{{validationFlag#GENERAL: Set True for validation prod - will create histograms#False}}"

pr.configName = "{{BKConfigName#GENERAL: BK configuration name e.g. MC #MC}}"
extraOptions = "{{extraOptions#GENERAL: extra options as python dict stepID:options#}}"

targets = "{{Target#PROD-1:MC: Target for MC (e.g. Tier2, ALL, LCG.CERN.cern or BAN:site1:site2#ALL}}"
eventsPerJob = "{{eventsPerJob#PROD-1:MC: Number of events per job#-1}}"
MCPriority = "{{MCPriority#PROD-1:MC: Production priority#0}}"
MCmulticoreFlag = "{{MCMulticoreFLag#PROD-1:MC multicore flag#True}}"
MCNumberOfProcessors = "{{MCNumberOfProcessors#PROD-1:MC jobs min/max n of processors#0,0}}"
MCSimulationType = "{{MCSimulationType#PROD-1:MC: type of MCSimulation#MCSimulation}}"
pr.mcTesting = ast.literal_eval("{{MCTesting#PROD-1:MC: testing enabled#True}}")
simulationCompressionLvl = "{{simulationCompressionLvl#PROD-1: Compression level#LOW}}"

simulationOutputVisFlag = ast.literal_eval(
    '{{simulationOutputVisFlag#PROD-1: Simulation visib flag dictionary (one flag per step {"step":"Y|N"})#{} }}'
)
try:
    simVisFlagSpecial = "{{simulationOutputVisFlagSpecial"
    simVisFlagSpecial += "#PROD-1: Special Visibility flag of output files"
    simVisFlagSpecial += ' (a dictionary {"step n":{("FType":flag)}} )#}}'
    simulationOutputVisFlagSpecial = ast.literal_eval(simVisFlagSpecial)
except SyntaxError:
    simulationOutputVisFlagSpecial = {}

selectionPlugin = "{{selectionPlugin#PROD-2:Selection: plugin e.g. Standard, BySize#BySize}}"
selectionGroupSize = "{{selectionGroupSize#PROD-2:Selection: input files total size (we'll download)#5}}"
selectionPriority = "{{selectionPriority#PROD-2:Selection: Job Priority e.g. 8 by default#8}}"
selectionCPU = "{{selectionCPU#PROD-2:Selection: Max CPU time in secs#100000}}"
removeInputSelection = "{{removeInputSelection#PROD-2:Selection: remove inputs#True}}"
selmulticoreFlag = "{{selMulticoreFLag#PROD-2:Selection: multicore flag#True}}"
selNumberOfProcessors = "{{selNumberOfProcessors#PROD-2:Selection jobs min/max n of processors#0,0}}"
selectionCompressionLvl = "{{selectionCompressionLvl#PROD-2:Selection: Compression level#LOW}}"
selectionOutputVisFlag = ast.literal_eval(
    '{{selectionOutputVisFlag#PROD-2: Selection visibility flag dictionary ({"step n": "Y|N"})#{} }}'
)
try:
    selVisFlagSpecial = "{{selectionOutputVisFlagSpecial"
    selVisFlagSpecial += '#PROD-2: Special Visibility flag of output files (a dictionary {"step n":{"FType":flag}} )#}}'
    selectionOutputVisFlagSpecial = ast.literal_eval(selVisFlagSpecial)
except SyntaxError:
    selectionOutputVisFlagSpecial = {}

mergingPlugin = "{{MergingPlugin#PROD-3:Merging: plugin e.g. Standard, BySize#BySize}}"
mergingGroupSize = "{{MergingGroupSize#PROD-3:Merging: Group Size e.g. BySize = GB file size#5}}"
mergingPriority = "{{MergingPriority#PROD-3:Merging: Job Priority e.g. 8 by default#8}}"
mergingCPU = "{{mergingCPU#PROD-3:Merging: Max CPU time in secs#100000}}"
removeInputMerge = "{{removeInputMerge#PROD-3:Merging: remove inputs#True}}"
mergemulticoreFlag = "{{mergeMulticoreFLag#PROD-3:Merging: multicore flag#True}}"
mergeNumberOfProcessors = "{{mergeNumberOfProcessors#PROD-3:Merging jobs min/max n of processors#0,0}}"
mergeCompressionLvl = "{{mergeCompressionLvl#PROD-3:Merging: Compression level#HIGH}}"
mergeOutputVisFlag = ast.literal_eval(
    '{{mergeOutputVisFlag#PROD-3: Merge visibility flag dictionary ({"step":"Y|N"})#{} }}'
)
try:
    mergeVisFlagSpecial = "{{mergeOutputVisFlagSpecial"
    mergeVisFlagSpecial += "#PROD-3: Special Visibility flag of output files"
    mergeVisFlagSpecial += '(a dictionary {"step n":{"FType":flag}} )#}}'
    mergeOutputVisFlagSpecial = ast.literal_eval(mergeVisFlagSpecial)
except SyntaxError:
    mergeOutputVisFlagSpecial = {}

pr.configVersion = "{{mcConfigVersion}}"
pr.eventType = "{{eventType}}"
# Often MC requests are defined with many subrequests but we want to retain
# the parent ID for viewing on the production monitoring page. If a parent
# request is defined then this is used.
pr.parentRequestID = "{{_parent}}"
pr.requestID = "{{ID}}"

if extraOptions:
    pr.extraOptions = ast.literal_eval(extraOptions)
pr.prodGroup = "{{pDsc}}"
pr.dataTakingConditions = "{{simDesc}}"

MCPriority = int(MCPriority)
selectionPriority = int(selectionPriority)
mergingPriority = int(mergingPriority)

removeInputMerge = ast.literal_eval(removeInputMerge)
removeInputSelection = ast.literal_eval(removeInputSelection)

MCNumberOfProcessors = tuple(int(x) for x in MCNumberOfProcessors.replace(" ", "").split(","))
selNumberOfProcessors = tuple(int(x) for x in selNumberOfProcessors.replace(" ", "").split(","))
mergeNumberOfProcessors = tuple(int(x) for x in mergeNumberOfProcessors.replace(" ", "").split(","))

###########################################
# LHCb conventions implied by the above
###########################################

localTestFlag = ast.literal_eval(localTestFlag)
validationFlag = ast.literal_eval(validationFlag)

if localTestFlag:
    pr.testFlag = True
    pr.publishFlag = False
    pr.prodsToLaunch = [1]

pr.outConfigName = pr.configName

w1 = ast.literal_eval(w1)
w2 = ast.literal_eval(w2)
w3 = ast.literal_eval(w3)

if not w1 and not w2 and not w3:
    gLogger.error("Vladimir, I told you to select at least one workflow!")
    DIRACexit(2)

elif w1:
    pr.prodsTypeList = [MCSimulationType]
    pr.outputSEs = ["Tier1_MC-DST"]

    pr.stepsInProds = [list(range(1, len(pr.stepsList) + 1))]
    pr.removeInputsFlags = [False]
    pr.priorities = [MCPriority]
    pr.cpus = [100000]
    pr.outputFileSteps = [str(len(pr.stepsList))]
    pr.targets = [targets]
    pr.events = [eventsPerJob]
    pr.groupSizes = [1]
    pr.plugins = [""]
    pr.inputDataPolicies = [""]
    pr.bkQueries = [""]
    pr.multicore = [MCmulticoreFlag]
    pr.processors = [MCNumberOfProcessors]

    simulationOutputVisFlag = fillVisList(simulationOutputVisFlag, len(pr.stepsInProds[0]))
    pr.compressionLvl = modifyCompression(
        simulationOutputVisFlag, simulationOutputVisFlagSpecial, simulationCompressionLvl
    )

    pr.outputVisFlag = [simulationOutputVisFlag]
    pr.specialOutputVisFlag = [simulationOutputVisFlagSpecial]

    # pr.resolveSteps()

elif w2:
    pr.prodsTypeList = [MCSimulationType, "MCReconstruction"]
    pr.outputSEs = ["Tier1-Buffer", "Tier1_MC-DST"]

    pr.stepsInProds = [
        [
            1,
        ],
        list(range(2, len(pr.stepsList) + 1)),
    ]
    pr.outputFileSteps = [str(len(pr.stepsInProds[0])), str(len(pr.stepsInProds[1]))]

    pr.removeInputsFlags = [False, removeInputSelection]
    pr.priorities = [MCPriority, selectionPriority]
    pr.cpus = [100000, selectionCPU]
    pr.targets = [targets, ""]
    pr.events = [eventsPerJob, -1]
    pr.groupSizes = [1, selectionGroupSize]
    pr.plugins = ["", selectionPlugin]
    pr.inputDataPolicies = ["", "download"]
    pr.bkQueries = ["", "fromPreviousProd"]
    pr.multicore = [MCmulticoreFlag, selmulticoreFlag]
    pr.processors = [MCNumberOfProcessors, selNumberOfProcessors]

    pr.compressionLvl = [simulationCompressionLvl] * len(pr.stepsInProds[0]) + [selectionCompressionLvl] * len(
        pr.stepsInProds[1]
    )

    simulationOutputVisFlag = fillVisList(simulationOutputVisFlag, len(pr.stepsInProds[0]))
    selectionOutputVisFlag = fillVisList(selectionOutputVisFlag, len(pr.stepsInProds[1]))

    pr.compressionLvl = modifyCompression(
        simulationOutputVisFlag, simulationOutputVisFlagSpecial, simulationCompressionLvl
    )
    pr.compressionLvl += modifyCompression(
        selectionOutputVisFlag, selectionOutputVisFlagSpecial, selectionCompressionLvl
    )

    pr.outputVisFlag = [simulationOutputVisFlag, selectionOutputVisFlag]
    pr.specialOutputVisFlag = [simulationOutputVisFlagSpecial, selectionOutputVisFlagSpecial]

    # pr.resolveSteps()

elif w3:
    pr.prodsTypeList = [MCSimulationType, "MCReconstruction", "MCMerge"]
    pr.outputSEs = ["Tier1-Buffer", "Tier1-Buffer", "Tier1_MC-DST"]

    pr.stepsInProds = [
        [
            1,
        ],
        list(range(2, len(pr.stepsList))),
        [len(pr.stepsList)],
    ]
    pr.outputFileSteps = ["1", str(len(pr.stepsInProds[1])), "1"]

    pr.removeInputsFlags = [False, removeInputSelection, removeInputMerge]
    pr.priorities = [MCPriority, selectionPriority, mergingPriority]
    pr.cpus = [100000, selectionCPU, mergingCPU]
    pr.targets = [targets, "", ""]
    pr.events = [eventsPerJob, -1, -1]
    pr.groupSizes = [1, selectionGroupSize, mergingGroupSize]
    pr.plugins = ["", selectionPlugin, mergingPlugin]
    pr.inputDataPolicies = ["", "download", "download"]
    pr.bkQueries = ["", "fromPreviousProd", "fromPreviousProd"]
    pr.multicore = [MCmulticoreFlag, selmulticoreFlag, mergemulticoreFlag]
    pr.processors = [MCNumberOfProcessors, selNumberOfProcessors, mergeNumberOfProcessors]

    simulationOutputVisFlag = fillVisList(simulationOutputVisFlag, len(pr.stepsInProds[0]))
    selectionOutputVisFlag = fillVisList(selectionOutputVisFlag, len(pr.stepsInProds[1]))
    mergeOutputVisFlag = fillVisList(mergeOutputVisFlag, len(pr.stepsInProds[2]))

    pr.compressionLvl = modifyCompression(
        simulationOutputVisFlag, simulationOutputVisFlagSpecial, simulationCompressionLvl
    )
    pr.compressionLvl += modifyCompression(
        selectionOutputVisFlag, selectionOutputVisFlagSpecial, selectionCompressionLvl
    )
    pr.compressionLvl += modifyCompression(mergeOutputVisFlag, mergeOutputVisFlagSpecial, mergeCompressionLvl)

    pr.outputVisFlag = [simulationOutputVisFlag, selectionOutputVisFlag, mergeOutputVisFlag]
    pr.specialOutputVisFlag = [simulationOutputVisFlagSpecial, selectionOutputVisFlagSpecial, mergeOutputVisFlagSpecial]

# In case we want just to test, we publish in the certification/test part of the BKK
if currentSetup == "LHCb-Certification" or pr.testFlag:
    pr.outConfigName = "certification"
    pr.configVersion = "test"

if pr.testFlag:
    pr.extend = "10"
    mergingGroupSize = "1"
    MCCpu = "50000"
    pr.previousProdID = 0  # set this for, e.g., launching only merging

# Validation implies few things, like saving all the outputs, and adding a GAUSSHIST
if validationFlag:
    pr.resolveSteps()
    pr.outConfigName = "validation"
    # Adding GAUSSHIST to the list of outputs to be produced (by the first step, which is Gauss)
    if "GAUSSHIST" not in pr.stepsListDict[0]["fileTypesOut"]:
        pr.stepsListDict[0]["fileTypesOut"].append("GAUSSHIST")
    pr.outputFileSteps = [""] * len(pr.prodsTypeList)


res = pr.buildAndLaunchRequest()
if not res["OK"]:
    gLogger.error(f"Errors with submission: {res['Message']}")
    DIRACexit(2)
else:
    gLogger.always(f"Submitted {str(res['Value'])}")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Simple extension of base class."""

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.TransformationSystem.Agent.ValidateOutputDataAgent import (
    ValidateOutputDataAgent as DIRACValidateOutputDataAgent,
)

from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient
from LHCbDIRAC.DataManagementSystem.Client.DataIntegrityClient import DataIntegrityClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

AGENT_NAME = "Transformation/ValidateOutputDataAgent"


class ValidateOutputDataAgent(DIRACValidateOutputDataAgent):
    """Simple extension of base class."""

    def __init__(self, *args, **kwargs):
        """c'tor."""
        DIRACValidateOutputDataAgent.__init__(self, *args, **kwargs)

        self.integrityClient = None
        self.fileCatalog = None
        self.transClient = None
        self.storageUsageClient = None

    def initialize(self):
        """standard initialize method for DIRAC agents."""
        res = DIRACValidateOutputDataAgent.initialize(self)
        if not res["OK"]:
            return res

        self.integrityClient = DataIntegrityClient()
        self.fileCatalog = FileCatalog()
        self.transClient = TransformationClient()
        self.storageUsageClient = StorageUsageClient()

        return S_OK()

    def checkTransformationIntegrity(self, prodID):
        """This method contains the real work."""
        gLogger.info("-" * 40)
        gLogger.info(f"Checking the integrity of production {prodID}")
        gLogger.info("-" * 40)

        res = self.getTransformationDirectories(prodID)
        if not res["OK"]:
            return res
        directories = res["Value"]

        # Skip debug directories
        directories = [d for d in directories if "debug" not in d]

        ######################################################
        #
        # This check performs BK->Catalog->SE
        #
        res = self.integrityClient.productionToCatalog(prodID)
        if not res["OK"]:
            gLogger.error(res["Message"])
            return res
        bk2catalogMetadata = res["Value"]["CatalogMetadata"]
        bk2catalogReplicas = res["Value"]["CatalogReplicas"]
        res = self.integrityClient.checkPhysicalFiles(bk2catalogReplicas, bk2catalogMetadata)
        if not res["OK"]:
            gLogger.error(res["Message"])
            return res

        if not directories:
            return S_OK()

        ######################################################
        #
        # This check performs Catalog->BK and Catalog->SE for possible output directories
        #
        res = self.fileCatalog.exists(directories)
        if not res["OK"]:
            gLogger.error(res["Message"])
            return res
        for directory, error in res["Value"]["Failed"]:
            gLogger.error("Failed to determine existance of directory", f"{directory} {error}")
        if res["Value"]["Failed"]:
            return S_ERROR("Failed to determine the existance of directories")
        directoryExists = res["Value"]["Successful"]
        for directory in sorted(directoryExists):
            if not directoryExists[directory]:
                continue
            iRes = self.integrityClient.catalogDirectoryToBK(directory)
            if not iRes["OK"]:
                gLogger.error(iRes["Message"])
                return iRes
            catalogDirMetadata = iRes["Value"]["CatalogMetadata"]
            catalogDirReplicas = iRes["Value"]["CatalogReplicas"]
            catalogMetadata = {}
            catalogReplicas = {}
            for lfn in catalogDirMetadata:
                if lfn not in bk2catalogMetadata:
                    catalogMetadata[lfn] = catalogDirMetadata[lfn]
                    if lfn in catalogDirReplicas:
                        catalogReplicas[lfn] = catalogDirReplicas[lfn]
            if not catalogMetadata:
                continue
            res = self.integrityClient.checkPhysicalFiles(catalogReplicas, catalogMetadata)
            if not res["OK"]:
                gLogger.error(res["Message"])
                return res

        return S_OK()

    def getTransformationDirectories(self, transID):
        """get the directories for the supplied transformation from the
        transformation system.

        :param self: self reference
        :param int transID: transformation ID
        """

        res = DIRACValidateOutputDataAgent.getTransformationDirectories(self, transID)

        if not res["OK"]:
            return res

        directories = res["Value"]

        if "StorageUsage" in self.directoryLocations:
            res = self.storageUsageClient.getStorageDirectories("", "", transID, [])
            if not res["OK"]:
                self.log.error("Failed to obtain storage usage directories", res["Message"])
                return res
            transDirectories = res["Value"]
            directories = self._addDirs(transID, transDirectories, directories)

        if not directories:
            self.log.info("No output directories found")
        directories = sorted(directories)
        return S_OK(directories)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set the Agent Type for a(the) given transformation(s)"""
import DIRAC

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} <Production ID> |<Production ID>",
                "Arguments:",
                "  <Production ID>:      DIRAC Production Id",
            ]
        )
    )

    Script.registerSwitch("a", "automatic", "Set automatic agent type")
    Script.registerSwitch("m", "manual", "Set manual agent type")

    Script.parseCommandLine(ignoreErrors=True)

    args = Script.getPositionalArgs()
    if len(args) < 1:
        Script.showHelp(exitCode=2)

    from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

    diracProd = DiracProduction()
    exitCode = 0
    errorList = []
    command = args[0]

    automatic = False
    manual = False
    type = ""

    switches = Script.getUnprocessedSwitches()

    for switch in switches:
        opt = switch[0].lower()

        if opt in ("a", "automatic"):
            automatic = True
        if opt in ("m", "manual"):
            manual = True

    if automatic and manual:
        print("ERROR: decide if you want automatic or manual ( not both ).")
        DIRAC.exit(2)
    elif not (automatic or manual):
        print("ERROR: decide if you want automatic or manual.")
        DIRAC.exit(2)
    elif automatic:
        type = "automatic"
    elif manual:
        type = "manual"

    for prodID in args:
        result = diracProd.production(prodID, type, disableCheck=False)
        if "Message" in result:
            errorList.append((prodID, result["Message"]))
            exitCode = 2
        elif not result:
            errorList.append((prodID, "Null result for production() call"))
            exitCode = 2
        else:
            exitCode = 0

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()
